#!/usr/bin/env python3
import argparse
import sys

from dp3 import worker


def init_parser(parser):
    parser.add_argument(
        "app_name",
        metavar="APP_NAME",
        help="Name of the application to distinct it from other DP3-based apps "
        "(it's used as a prefix of RabbitMQ queue names, for example).",
    )
    parser.add_argument(
        "config_dir",
        metavar="CONFIG_DIRECTORY",
        help="Path to a directory containing configuration files (e.g. /etc/my_app/config)",
    )
    parser.add_argument(
        "process_index",
        metavar="PROCESS_INDEX",
        type=int,
        help="Index of this worker process. For each application there must be N processes running "
        "simultaneously, each started with a unique index (from 0 to N-1). N is read from "
        "configuration ('worker_processes' in 'processing_core.yml').",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="More verbose output (set log level to DEBUG).",
    )


def run():
    print(
        "WARNING: The `worker` entrypoint is deprecated due to possible namespace conflicts. "
        "Please use `dp3 worker` instead.",
        file=sys.stderr,
    )

    # Parse arguments
    parser = argparse.ArgumentParser(
        prog="worker",
        description="Main worker process of the DP3 platform. "
        "There are usually multiple workers running in parallel.",
    )
    init_parser(parser)
    args = parser.parse_args()

    main(args)


def main(args):
    worker.main(args.app_name, args.config_dir, args.process_index, args.verbose)


if __name__ == "__main__":
    run()



from datetime import datetime
from ipaddress import IPv4Address, IPv6Address
from typing import Annotated, Any, Optional, Union

from pydantic import BaseModel, BeforeValidator, Field, PlainSerializer

from dp3.common.types import T2Datetime


def to_json_friendly(v):
    if isinstance(v, (IPv4Address, IPv6Address)):
        return str(v)
    return v


def ignore_value(_v):
    """Ignore the passed value and return None."""
    return None


class DataPointBase(BaseModel, use_enum_values=True):
    """Data-point

    Contains single raw data value received on API.
    This is just base class - plain, observation or timeseries datapoints inherit from this class
    (see below).

    Provides front line of validation for this data value.

    Internal usage: inside Task, created by TaskExecutor
    """

    etype: str
    eid: str
    attr: str
    src: Optional[str] = None
    v: Annotated[Any, PlainSerializer(to_json_friendly, when_used="json")] = None
    c: Any = None
    t1: Any = None
    t2: Any = None


class DataPointPlainBase(DataPointBase):
    """Plain attribute data-point

    Contains single raw data value received on API for plain attribute.

    In case of plain data-point, it's not really a data-point, but we use
    the same naming for simplicity.
    """

    t1: Annotated[None, BeforeValidator(ignore_value)] = None
    t2: None = None


class DataPointObservationsBase(DataPointBase):
    """Observations attribute data-point

    Contains single raw data value received on API for observations attribute.
    """

    t1: datetime
    t2: T2Datetime = Field(None, validate_default=True)
    c: Annotated[float, Field(ge=0.0, le=1.0)] = 1.0


class DataPointTimeseriesBase(DataPointBase):
    """Timeseries attribute data-point

    Contains single raw data value received on API for observations attribute.
    """

    t1: datetime
    t2: T2Datetime = Field(None, validate_default=True)


def is_list_ordered(to_check: list):
    """Checks if list is ordered (not decreasing anywhere)"""
    return all(to_check[i] <= to_check[i + 1] for i in range(len(to_check) - 1))


# Timeseries validators
# They are included by child classes of `DataPointTimeseriesBase` in `AttrSpecTimeseries`,
# where `v` is available
# I know, this is a bit dirty, but there's no other way around.
def dp_ts_v_validator(v):
    """Check if all value arrays are the same length."""
    values_len = {len(v_i) for _, v_i in v.model_dump().items()}
    assert len(values_len) == 1, f"Series values have different lengths: {values_len}"

    return v


def dp_ts_root_validator_regular_wrapper(time_step):
    def dp_ts_root_validator_regular(self):
        """Validates or sets t2 of regular timeseries datapoint"""
        # Get length of first value series included in datapoint
        first_series = list(self.v.model_dump().keys())[0]
        series_len = len(self.v.model_dump()[first_series])
        correct_t2 = self.t1 + series_len * time_step

        if self.t2:
            assert (
                self.t2 == correct_t2
            ), "Difference of t1 and t2 is invalid. Must be values_len*time_step."
        else:
            self.t2 = correct_t2

        return self

    return dp_ts_root_validator_regular


def dp_ts_root_validator_irregular(self):
    """Validates or sets t2 of irregular timeseries datapoint"""
    first_time = self.v.time[0]
    last_time = self.v.time[-1]

    # Check t1 <= first_time
    assert self.t1 <= first_time, f"'t1' is above first item in 'time' series ({first_time})"

    # Check last_time <= t2
    if self.t2:
        assert self.t2 >= last_time, f"'t2' is below last item in 'time' series ({last_time})"
    else:
        self.t2 = last_time

    # time must be ordered
    assert is_list_ordered(self.v.time), "'time' series is not ordered"

    return self


def dp_ts_root_validator_irregular_intervals(self):
    """Validates or sets t2 of irregular intervals timeseries datapoint"""
    first_time = self.v.time_first[0]
    last_time = self.v.time_last[-1]

    # Check t1 <= first_time
    assert self.t1 <= first_time, f"'t1' is above first item in 'time_first' series ({first_time})"

    # Check last_time <= t2
    if self.t2:
        assert self.t2 >= last_time, f"'t2' is below last item in 'time_last' series ({last_time})"
    else:
        self.t2 = last_time

    # Check time_first[i] <= time_last[i]
    assert all(
        t[0] <= t[1] for t in zip(self.v.time_first, self.v.time_last)
    ), "'time_first[i] <= time_last[i]' isn't true for all 'i'"

    return self


DataPointType = Union[DataPointPlainBase, DataPointObservationsBase, DataPointTimeseriesBase]



"""
Platfrom HTTP API implementation.

See the individual routers to see the API endpoint implementation.

- [`root`][dp3.api.routers.root] - catch-all router for the root path `/*`.
  Implements the datapoint ingestion endpoint.
- [`entity`][dp3.api.routers.entity] - implements the entity endpoints, `/{entity_type}/*`.
- [`control`][dp3.api.routers.control] - implements the control endpoints, `/{control_action}`.
"""



import logging

from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig
from dp3.common.state import SharedFlag


class BaseModule:
    """
    Base class for platform modules.
    Every module must inherit this abstract class for automatic loading of module!

    Attributes:
        refresh: SharedFlag that is set to True when module should refresh its attributes values
        log: Logger for the module
    """

    def __init__(
        self, platform_config: PlatformConfig, module_config: dict, registrar: CallbackRegistrar
    ):
        """Initialize the module and register callbacks.

        `self.load_config()` is called in the base class.

        Args:
            platform_config: Platform configuration class
            module_config: Configuration of the module,
                equivalent of `platform_config.config.get("modules.<module_name>")`
            registrar: A callback / hook registration interface
        """
        self.refresh: SharedFlag = SharedFlag(False, banner=f"Refresh {self.__class__.__name__}")
        self.log: logging.Logger = logging.getLogger(self.__class__.__name__)

        self.load_config(platform_config, module_config)

    def load_config(self, config: PlatformConfig, module_config: dict) -> None:
        """Load module configuration.

        Is called on module initialization and on refresh request via `/control` API.
        """

    def start(self) -> None:
        """
        Run the module - used to run own thread if needed.

        Called after initialization, may be used to create and run a separate
        thread if needed by the module. Do nothing unless overridden.
        """
        return None

    def stop(self) -> None:
        """
        Stop the module - used to stop own thread.

        Called before program exit, may be used to finalize and stop the
        separate thread if it is used. Do nothing unless overridden.
        """
        return None



from datetime import datetime
from typing import Annotated, Any, Optional

from pydantic import BaseModel, Field, NonNegativeInt, RootModel

from dp3.common.attrspec import AttrSpecType, AttrType


class EntityState(BaseModel):
    """Entity specification and current state

    Merges (some) data from DP3's `EntitySpec` and state information from `Database`.
    Provides estimate count of master records in database.
    """

    id: str
    name: str
    attribs: dict[str, Annotated[AttrSpecType, Field(discriminator="type")]]
    eid_estimate_count: NonNegativeInt


class EntityEidList(BaseModel):
    """List of entity eids and their data based on latest snapshot

    Includes timestamp of latest snapshot creation, count of returned documents
    and total count of documents available under specified filter.

    Data does not include history of observations attributes and timeseries.
    """

    time_created: Optional[datetime] = None
    count: int
    total_count: int
    data: list[dict]


EntityEidMasterRecord = RootModel[dict]

EntityEidSnapshots = RootModel[list[dict]]


class EntityEidData(BaseModel):
    """Data of entity eid

    Includes all snapshots and master record.

    `empty` signalizes whether this eid includes any data.
    """

    empty: bool
    master_record: dict
    snapshots: list[dict]


class EntityEidAttrValueOrHistory(BaseModel):
    """Value and/or history of entity attribute for given eid

    Depends on attribute type:
    - plain: just (current) value
    - observations: (current) value and history stored in master record (optionally filtered)
    - timeseries: just history stored in master record (optionally filtered)
    """

    attr_type: AttrType
    current_value: Any = None
    history: list[dict] = []


class EntityEidAttrValue(BaseModel):
    """Value of entity attribute for given eid

    The value is fetched from master record.
    """

    value: Any = None



import json
import sys
from time import sleep

import common

from dp3.api.internal.entity_response_models import EntityEidList


class GetEntityEids(common.APITest):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        dp_base = {"src": "setup@test", "attr": "data1", "type": "A"}
        for i in range(0, 100, 20):
            res = cls.push_datapoints(
                [{**dp_base, "id": f"A{i}", "v": f"v{i}"} for i in range(i, i + 20)]
            )
            print(res.content.decode("utf-8"), file=sys.stderr)
        sleep(8)
        cls.get_request("control/make_snapshots")
        sleep(6)

    def test_get_entity_eids(self):
        eids = self.get_entity_data("entity/A", EntityEidList)
        self.assertEqual(20, len(eids.data))

    def test_get_entity_eids_pagination(self):
        expected_eids = {f"A{i}" for i in range(0, 100)}
        received_eids = set()

        for i in range(0, 100, 10):
            eids = self.get_entity_data("entity/A", EntityEidList, skip=i, limit=10)
            self.assertEqual(10, len(eids.data), f"Failed at {i}")
            received_eids.update(x["eid"] for x in eids.data)

        eids = self.get_entity_data("entity/A", EntityEidList, skip=101, limit=20)
        self.assertEqual(0, len(eids.data))
        self.assertSetEqual(expected_eids, received_eids)

    def test_get_entity_eids_generic_filter(self):
        eids = self.get_entity_data(
            "entity/A", EntityEidList, generic_filter=json.dumps({"last.eid": "A0"})
        )
        self.assertEqual(1, len(eids.data))
        self.assertEqual("A0", eids.data[0]["eid"])

    def test_get_entity_eids_fulltext_filters_eid(self):
        eids = self.get_entity_data(
            "entity/A", EntityEidList, fulltext_filters=json.dumps({"eid": "A5.*"})
        )
        self.assertEqual(11, len(eids.data))  # A5, A50 ... A59



"""DP3 Setup Script for creating a DP3 application."""

import argparse
import contextlib
import shutil
import sys
from pathlib import Path


def replace_template(directory: Path, template: str, replace_with: str):
    """Replace all occurrences of `template` with the given text."""
    for file in directory.rglob("*"):
        if file.is_file():
            replace_template_file(file, template, replace_with)


def replace_template_file(file: Path, template: str, replace_with: str):
    with contextlib.suppress(UnicodeDecodeError, PermissionError), file.open("r+") as f:
        contents = f.read()
        contents = contents.replace(template, replace_with)
        f.seek(0)
        f.write(contents)
        f.truncate()


def init_parser(parser):
    """
    Initialize an argparse object to parse the project directory and the app name.
    """
    parser.add_argument("project_dir", help="The project directory.", type=str)
    parser.add_argument("app_name", help="The name of the application.", type=str)


def run():
    print(
        "WARNING: The `dp3-setup` entrypoint is deprecated. Please use `dp3 setup` instead. ",
        file=sys.stderr,
    )

    parser = argparse.ArgumentParser(description="Create a DP3 application.")
    init_parser(parser)
    args = parser.parse_args()

    main(args)


def main(args):
    app_name = args.app_name.lower()

    # Ensure the project directory exists.
    project_dir = Path(args.project_dir)
    project_dir = project_dir.absolute()
    project_dir.mkdir(exist_ok=True, parents=True)

    # Get the current package location.
    package_dir = Path(__file__).parent.parent

    # Copy the template files to the project directory.
    shutil.copytree(package_dir / "template" / "app", project_dir, dirs_exist_ok=True)

    replace_template(project_dir, "{{DP3_APP}}", app_name)
    replace_template(project_dir, "__dp3_app__", app_name)


if __name__ == "__main__":
    run()



"""DP3 Setup Config Script for creating a DP3 application."""

import shutil
import sys
from pathlib import Path

from dp3.bin.setup import replace_template, replace_template_file
from dp3.common.config import read_config_dir


def init_parser(parser):
    """
    There are two desired use-cases for this command:

    ```
    dp3 config nginx --app-name <APP_NAME> --hostname <SERVER_HOSTNAME> --www-root <DIRECTORY>
    ```

    ```
    dp3 config supervisor --app-name <APP_NAME> --config <CONFIG_DIR>
    ```
    """
    parser.add_argument(
        "subcommand",
        help="What configuration to setup. "
        "'nginx' requires appname, hostname and www_root, "
        "'supervisor' requires appname and config.",
        choices=["nginx", "supervisor"],
    )
    parser.add_argument(
        "--app-name", dest="app_name", help="The name of the application.", type=str
    )
    parser.add_argument(
        "--hostname", dest="server_hostname", help="The hostname of the server.", type=str
    )
    parser.add_argument(
        "--www-root",
        dest="www_root",
        help="The directory where served HTML content will be placed.",
        type=str,
    )
    parser.add_argument(
        "--config",
        dest="config_dir",
        help="The directory where the DP3 config of your application is stored.",
        type=str,
    )


def config_nginx(app_name, server_hostname, www_root):
    """Configure nginx to serve the application."""
    # Get the current package location.
    package_dir = Path(__file__).parent.parent

    nginx_dir = Path("/etc/nginx/")

    # Copy the template files to the project directory.
    shutil.copytree(package_dir / "template" / "nginx", nginx_dir, dirs_exist_ok=True)

    replace_template(nginx_dir, "{{DP3_APP}}", app_name)
    replace_template(nginx_dir, "__SERVER_NAME__", server_hostname)
    replace_template(nginx_dir, "__WWW_ROOT__", www_root)

    # Create the www root directory.
    www_root_dir = Path(www_root)
    www_root_dir.mkdir(exist_ok=True, parents=True)
    shutil.copytree(package_dir / "template" / "html", www_root_dir, dirs_exist_ok=True)

    replace_template(www_root_dir, "{{DP3_APP}}", app_name)
    replace_template(www_root_dir, "{{HOSTNAME}}", server_hostname)

    shutil.chown(www_root_dir, user=app_name, group=app_name)
    www_root_dir.chmod(0o775)


def get_python_directories() -> tuple[Path, Path]:
    """
    Get the directory where the DP3 executable is located and where the DP3 library is installed.

    There are two cases we need to consider:

    - The package is installed in a normal way, be it in virtual environment,
      in `~/.local` or in `/usr/local`. The path will allways look something like
      `/path/to/python/lib/python3.X/{site|dist}-packages/dp3`.
      The path to the executable will be `/path/to/python/bin/dp3`.
    - The package is installed in editable mode, in which case the path of the
      package depends on the location of the repo. The path to the executable relates
      in no way to the path of the package, we cannot be sure.
      We therefore give it our best guess, which is the path of the python executable.
      This will be correct only in the case of a virtual environment, but it is the best
      we can do.
    """
    package_path = Path(__file__).parent.parent.absolute()
    packages_path = package_path.parent
    lib_python_path = packages_path.parent

    if "-packages" in packages_path.name and "python3" in lib_python_path.name:
        binary_path = package_path.parent.parent.parent.parent / "bin"
    else:  # This is a development install.
        binary_path = Path(sys.executable).parent.absolute()
    return binary_path, package_path


def config_supervisor(app_name, config_dir):
    """
    Configure supervisor for process management.
    Replaces templates: DP3_EXE, DP3_APP, CONFIG_DIR, DP3_BIN, DP3_LIB
    """
    # Get the current package location and other relative directories.
    python_bin_dir, package_dir = get_python_directories()
    dp3_executable_path = str(python_bin_dir / "dp3")
    abs_config_dir = Path(config_dir).absolute()

    # Load configured worker count from dp3 config.
    config = read_config_dir(config_dir, recursive=True)
    worker_count = int(config.get("processing_core.worker_processes"))

    # Ensure the config directory exists.
    supervisor_dir = Path(f"/etc/{app_name}/")
    supervisor_dir.mkdir(exist_ok=True, parents=True)
    shutil.chown(supervisor_dir, user=app_name, group=app_name)
    supervisor_dir.chmod(0o775)

    # Copy the template files to the project directory.
    shutil.copytree(package_dir / "template" / "supervisor", supervisor_dir, dirs_exist_ok=True)

    replace_template(supervisor_dir, "{{DP3_APP}}", app_name)
    replace_template(supervisor_dir, "{{CONFIG_DIR}}", str(abs_config_dir))
    replace_template(supervisor_dir, "{{DP3_BIN}}", str(python_bin_dir))
    replace_template(supervisor_dir, "{{DP3_EXE}}", dp3_executable_path)
    replace_template(supervisor_dir, "{{DP3_PACKAGE_DIR}}", str(package_dir))
    replace_template(supervisor_dir, "{{WORKER_COUNT}}", str(worker_count))

    # Set up the systemd service to start supervisor.
    service_path = Path(f"/etc/systemd/system/{app_name}.service")
    shutil.copy(
        package_dir / "template" / "systemd" / "app.service",
        service_path,
    )
    replace_template_file(service_path, "{{DP3_APP}}", app_name)

    # Set up the appctl script.
    appctl_path = Path(f"/usr/bin/{app_name}ctl")
    shutil.copy(
        package_dir / "template" / "appctl",
        appctl_path,
    )
    replace_template_file(appctl_path, "{{DP3_APP}}", app_name)
    appctl_path.chmod(0o755)


def main(args):
    required_args = []
    if args.subcommand == "nginx":
        required_args = ["app_name", "server_hostname", "www_root"]
    if args.subcommand == "supervisor":
        required_args = ["app_name", "config_dir"]
    for arg in required_args:
        if not getattr(args, arg):
            print(
                f"Missing required argument: {arg}. "
                f"The {args.subcommand} requires the following arguments: {required_args}",
                file=sys.stderr,
            )
            sys.exit(1)

    if args.subcommand == "nginx":
        config_nginx(args.app_name, args.server_hostname, args.www_root)
    if args.subcommand == "supervisor":
        config_supervisor(args.app_name, args.config_dir)



from time import sleep

import common
from common import ACCEPTED_ERROR_CODES

from dp3.api.internal.entity_response_models import EntityEidAttrValueOrHistory

TESTED_PATH = "entity/{entity}/{eid}/get/{attr}"
TESTED_HISTORY_PATH = TESTED_PATH.format(
    entity="test_entity_type", eid="test_entity_id", attr="test_attr_history"
)
TESTED_INT_PATH = TESTED_PATH.format(
    entity="test_entity_type", eid="test_entity_id", attr="test_attr_int"
)


class GetEidAttrValue(common.APITest):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        sleep(10)

    def test_unknown_entity_type(self):
        response = self.get_request(
            TESTED_PATH.format(entity="xyz", eid="test_entity_id", attr="test_attr_int")
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_unknown_attr_name(self):
        response = self.get_request(
            TESTED_PATH.format(entity="test_entity_type", eid="test_entity_id", attr="xyz")
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_unknown_entity_id(self):
        payload = self.get_entity_data(
            TESTED_PATH.format(entity="test_entity_type", eid="xyz", attr="test_attr_int"),
            EntityEidAttrValueOrHistory,
        )
        expected = EntityEidAttrValueOrHistory(attr_type=1)
        self.assertEqual(payload, expected, "The returned object is not empty")

    def test_valid_get(self):
        response = self.push_datapoints(
            [
                {
                    "type": "test_entity_type",
                    "id": "test_entity_id",
                    "attr": "test_attr_int",
                    "v": 123,
                }
            ]
        )
        self.assertEqual(200, response.status_code, "setup push failed")
        expected = EntityEidAttrValueOrHistory(attr_type=1, current_value=123)
        self.query_expected_value(
            lambda: self.get_entity_data(TESTED_INT_PATH, EntityEidAttrValueOrHistory),
            lambda received: received == expected,
        )

    def test_invalid_timestamp(self):
        response = self.get_request(TESTED_HISTORY_PATH, date_from="xyz")
        self.assertEqual(422, response.status_code)

    def test_valid_time_intervals(self):
        time_intervals = {
            "<T1, T2>": {"date_from": "2020-01-01T00:00:00", "date_to": "2020-01-01T00:00:00"},
            "<T1, _>": {"date_from": "2020-01-01T00:00:00"},
            "<_, T2>": {"date_to": "2020-01-01T00:00:00"},
            "<_, _>": {},
        }
        for msg, intervals in time_intervals.items():
            with self.subTest(msg=msg):
                response = self.get_request(TESTED_HISTORY_PATH, **intervals)
                self.assertEqual(200, response.status_code)



import sys

import common
from common import ACCEPTED_ERROR_CODES


class PushDatapoints(common.APITest):
    def test_invalid_payload(self):
        response = self.push_datapoints(
            {"type": "test_entity_type", "attr": "test_attr_int", "id": "test_entity_id", "v": 123}
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES, "invalid payload (not a list)")

        response = self.push_datapoints(["xyz"])
        self.assertIn(
            response.status_code,
            ACCEPTED_ERROR_CODES,
            "invalid payload (list element is not a dictionary)",
        )

    def test_invalid_content_type(self):
        response = self.post_request("datapoints", data="bodyisstring")
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES, "invalid payload (not a list)")

    def test_unknown_entity_type(self):
        response = self.push_datapoints(
            [{"type": "xyz", "attr": "test_attr_int", "id": "test_entity_id", "v": 123}]
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_unknown_attr_name(self):
        response = self.push_datapoints(
            [{"type": "test_entity_type", "id": "test_entity_id", "attr": "xyz", "v": 123}]
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_invalid_timestamp(self):
        response = self.push_datapoints(
            [
                {
                    "type": "test_entity_type",
                    "id": "test_entity_id",
                    "attr": "test_attr_history",
                    "v": 123,
                    "t1": "xyz",
                }
            ]
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def test_missing_value(self):
        response = self.push_datapoints(
            [{"type": "test_entity_type", "id": "test_entity_id", "attr": "test_attr_int"}]
        )
        self.assertIn(response.status_code, ACCEPTED_ERROR_CODES)

    def helper_test_datatype_value(self, data_type: str, value, expected_codes: set[int]):
        response = self.push_datapoints(
            [
                {
                    "type": "test_entity_type",
                    "id": "test_entity_id",
                    "attr": f"test_attr_{data_type}",
                    "v": value,
                }
            ]
        )
        print(response.content.decode("utf-8"), file=sys.stderr)
        self.assertIn(response.status_code, expected_codes)

    def test_data_type_values_valid(self):
        for data_type, valid in common.values["valid"].items():
            for value in valid:
                with self.subTest(data_type=data_type, v=value):
                    self.helper_test_datatype_value(data_type, value=value, expected_codes={200})

    def test_data_type_values_invalid(self):
        for data_type, valid in common.values["invalid"].items():
            for value in valid:
                with self.subTest(data_type=data_type, v=value):
                    self.helper_test_datatype_value(
                        data_type, value=value, expected_codes=ACCEPTED_ERROR_CODES
                    )



from datetime import datetime
from typing import Annotated, Any, Optional

from pydantic import BaseModel, Field, model_validator

from dp3.api.internal.helpers import api_to_dp3_datapoint
from dp3.common.types import T2Datetime


class DataPoint(BaseModel):
    """Data-point for API

    Contains single raw data value received on API.
    This is generic class for plain, observation and timeseries datapoints.

    Provides front line of validation for this data value.

    This differs slightly compared to `DataPoint` from DP3 in naming of attributes due to historic
    reasons.

    After validation of this schema, datapoint is validated using attribute-specific validator to
    ensure full compilance.
    """

    type: str
    id: str
    attr: str
    v: Any
    t1: Optional[datetime] = None
    t2: Optional[T2Datetime] = Field(None, validate_default=True)
    c: Annotated[float, Field(ge=0.0, le=1.0)] = 1.0
    src: Optional[str] = None

    @model_validator(mode="after")
    def validate_against_attribute(self):
        # Try to convert API datapoint to DP3 datapoint
        try:
            api_to_dp3_datapoint(self.model_dump())
        except KeyError as e:
            raise ValueError(f"Missing key: {e}") from e

        return self



import datetime
from time import sleep

import common

from dp3.api.internal.entity_response_models import EntityEidData


class SnapshotIntegration(common.APITest):
    @classmethod
    def setUpClass(cls) -> None:
        now = datetime.datetime.now()
        t1 = (now - datetime.timedelta(minutes=30)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
        t2 = (now + datetime.timedelta(minutes=30)).strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]

        entity_datapoints = [
            # For test_linked_entities_loaded
            {"type": "B", "id": "b1", "attr": "as", "v": "a1", "t1": t1, "t2": t2, "c": 1.0},
            {"type": "B", "id": "b1", "attr": "data1", "v": "initb"},
            {"type": "B", "id": "b1", "attr": "data2", "v": "initb"},
            {"type": "A", "id": "a1", "attr": "bs", "v": "b1", "t1": t1, "t2": t2, "c": 1.0},
            {"type": "A", "id": "a1", "attr": "data1", "v": "inita"},
            {"type": "A", "id": "a1", "attr": "data2", "v": "inita"},
            # For test_weak_component_detection
            {"type": "B", "id": "b2", "attr": "as", "v": "a1", "t1": t1, "t2": t2, "c": 1.0},
            {"type": "B", "id": "b2", "attr": "data1", "v": "initb"},
            {"type": "B", "id": "b2", "attr": "data2", "v": "initb"},
            {"type": "A", "id": "a2", "attr": "data1", "v": "inita"},
            {"type": "A", "id": "a2", "attr": "data2", "v": "inita"},
            # For test_hook_dependency_value_forwarding
            {"type": "D", "id": "d1", "attr": "cs", "v": "c1", "t1": t1, "t2": t2, "c": 1.0},
            {"type": "D", "id": "d1", "attr": "data1", "v": "initb"},
            {"type": "D", "id": "d1", "attr": "data2", "v": "initb"},
            {"type": "C", "id": "c1", "attr": "ds", "v": "d1", "t1": t1, "t2": t2, "c": 1.0},
            {"type": "C", "id": "c1", "attr": "data1", "v": "inita"},
            {"type": "C", "id": "c1", "attr": "data2", "v": "inita"},
        ]
        cls.push_datapoints(entity_datapoints)
        cls.get_request("control/make_snapshots")
        sleep(3)

    def test_linked_entities_loaded(self):
        for entity, eid in [("A", "a1"), ("B", "b1")]:
            data = self.get_entity_data(f"entity/{entity}/{eid}", EntityEidData)
            for snapshot in data.snapshots:
                self.assertEqual(snapshot["data1"], "initb")
                self.assertEqual(snapshot["data2"], "inita")

    def test_weak_component_detection(self):
        """
        Remove link from the first entity to be processed, so that an elementary BFS starting
        from this entity is not enough to load the entire weakly connected component.
        """
        for entity, eid in [("A", "a2"), ("B", "b2")]:
            data = self.get_entity_data(f"entity/{entity}/{eid}", EntityEidData)
            for snapshot in data.snapshots:
                self.assertEqual(snapshot["data1"], "initb")
                self.assertEqual(snapshot["data2"], "inita")

    def test_hook_dependency_value_forwarding(self):
        for entity, eid in [("D", "d1"), ("C", "c1")]:
            data = self.get_entity_data(f"entity/{entity}/{eid}", EntityEidData)
            for snapshot in data.snapshots:
                self.assertEqual(snapshot["data1"], "modifd")
                self.assertEqual(snapshot["data2"], "modifc")



from datetime import timedelta
from enum import Flag
from typing import Annotated, Any, Literal, Optional, Union

from pydantic import (
    BaseModel,
    Field,
    PositiveInt,
    PrivateAttr,
    create_model,
    field_validator,
    model_validator,
)
from pydantic_core.core_schema import FieldValidationInfo

from dp3.common.datapoint import (
    DataPointBase,
    DataPointObservationsBase,
    DataPointPlainBase,
    DataPointTimeseriesBase,
    dp_ts_root_validator_irregular,
    dp_ts_root_validator_irregular_intervals,
    dp_ts_root_validator_regular_wrapper,
    dp_ts_v_validator,
)
from dp3.common.datatype import DataType, ReadOnly
from dp3.common.entityspec import SpecModel
from dp3.common.types import ParsedTimedelta

# Regex of attribute and series id's
ID_REGEX = r"^[a-zA-Z_][a-zA-Z0-9_-]*$"

# Dict of timeseries type spec
timeseries_types = {
    "regular": {"default_series": {}, "sort_by": "t1"},
    "irregular": {"default_series": {"time": {"data_type": "time"}}, "sort_by": "time"},
    "irregular_intervals": {
        "default_series": {"time_first": {"data_type": "time"}, "time_last": {"data_type": "time"}},
        "sort_by": "time_first",
    },
}


class AttrType(Flag):
    """Enum of attribute types

    `PLAIN` = 1
    `OBSERVATIONS` = 2
    `TIMESERIES` = 4
    """

    PLAIN = 1
    OBSERVATIONS = 2
    TIMESERIES = 4

    @classmethod
    def from_str(cls, type_str: str):
        """
        Convert string representation like "plain" to AttrType.
        """
        try:
            return cls(cls[type_str.upper()])
        except Exception as e:
            raise ValueError(f"Invalid `type` of attribute '{type_str}'") from e


class ObservationsHistoryParams(BaseModel):
    """History parameters field of observations attribute"""

    max_age: Optional[ParsedTimedelta] = None
    max_items: Optional[PositiveInt] = None
    expire_time: Optional[ParsedTimedelta] = None
    pre_validity: Optional[ParsedTimedelta] = timedelta()
    post_validity: Optional[ParsedTimedelta] = timedelta()

    aggregate: bool = True

    @field_validator("expire_time", mode="before")
    @classmethod
    def expire_time_inf_transform(cls, v):
        return None if v == "inf" else v


class TimeseriesTSParams(BaseModel):
    """Timeseries parameters field of timeseries attribute"""

    max_age: Optional[ParsedTimedelta] = None
    time_step: Optional[ParsedTimedelta] = None


class TimeseriesSeries(BaseModel):
    """Series of timeseries attribute"""

    data_type: DataType

    @field_validator("data_type")
    @classmethod
    def check_series_data_type(cls, v):
        assert str(v) in [
            "int",
            "int64",
            "float",
            "time",
        ], f"Data type of series must be one of int, int64, float, time; not {v}"
        return v


class AttrSpecGeneric(SpecModel, use_enum_values=True):
    """Base of attribute specification

    Parent of other `AttrSpec` classes.

    Attributes:
        ttl: Optional extension of TTL of the entity
            - will be ignored if lifetime setting does not match.
    """

    id: str = Field(pattern=ID_REGEX)
    name: str
    description: str = ""
    ttl: Optional[ParsedTimedelta] = timedelta()

    _dp_model = PrivateAttr()

    @property
    def dp_model(self) -> DataPointBase:
        return self._dp_model


class AttrSpecClassic(AttrSpecGeneric):
    """Parent of non-timeseries `AttrSpec` classes."""

    data_type: DataType
    editable: bool = False

    @property
    def is_iterable(self) -> bool:
        """Returns whether specified attribute is iterable."""
        return self.data_type.iterable

    @property
    def element_type(self) -> DataType:
        """Returns the element type for iterable data types."""
        return self.data_type.elem_type

    @property
    def is_relation(self) -> bool:
        """Returns whether specified attribute is a link."""
        return self.data_type.is_link or (
            self.data_type.iterable and self.data_type.elem_type.is_link
        )

    @property
    def relation_to(self) -> str:
        """Returns linked entity id. Raises ValueError if attribute is not a link."""
        if self.data_type.iterable:
            return self.element_type.get_linked_entity()
        return self.data_type.get_linked_entity()

    @property
    def is_mirrored(self) -> bool:
        """Returns whether specified attribute is a mirrored link."""
        if self.data_type.iterable:
            return self.element_type.mirror_link
        return self.data_type.mirror_link

    @property
    def mirror_as(self) -> str:
        """
        Returns:
             name of the mirrored attribute.
        Raises:
            ValueError: If attribute is not a mirrored link.
        """
        if self.data_type.iterable:
            return self.element_type.mirror_as
        return self.data_type.mirror_as


class AttrSpecPlain(AttrSpecClassic):
    """Plain attribute specification"""

    t: AttrType = AttrType.PLAIN
    type: Literal["plain"] = Field(..., repr=False)

    def __init__(self, **data):
        super().__init__(**data)

        self._dp_model = create_model(
            f"DataPointPlain_{self.id}",
            __base__=DataPointPlainBase,
            v=(self.data_type.data_type, ...),
        )


class AttrSpecReadOnly(AttrSpecPlain):
    """Read-only plain attribute specification. Used for internal attributes."""

    def __init__(self, **data):
        super().__init__(**data)

        self._dp_model = create_model(
            f"DataPointReadOnly_{self.id}",
            __base__=DataPointPlainBase,
            v=(ReadOnly, ...),
        )


class AttrSpecObservations(AttrSpecClassic):
    """Observations attribute specification"""

    t: AttrType = AttrType.OBSERVATIONS
    type: Literal["observations"] = Field(..., repr=False)

    confidence: bool = False
    multi_value: bool = False
    history_params: ObservationsHistoryParams = Field(default_factory=ObservationsHistoryParams)
    history_force_graph: bool = False

    def __init__(self, **data):
        super().__init__(**data)

        value_validator = self.data_type.data_type

        self._dp_model = create_model(
            f"DataPointObservations_{self.id}",
            __base__=DataPointObservationsBase,
            v=(value_validator, ...),
        )


class AttrSpecTimeseries(AttrSpecGeneric):
    """Timeseries attribute specification"""

    t: AttrType = AttrType.TIMESERIES
    type: Literal["timeseries"] = Field(..., repr=False)

    timeseries_type: Literal["regular", "irregular", "irregular_intervals"]
    series: dict[Annotated[str, Field(pattern=ID_REGEX)], TimeseriesSeries] = {}
    timeseries_params: TimeseriesTSParams

    def __init__(self, **data):
        super().__init__(**data)

        # Typing of `v` field
        dp_value_typing = {}
        for s in self.series:
            data_type = self.series[s].data_type.data_type
            dp_value_typing[s] = ((list[data_type]), ...)

        # Add root validator
        if self.timeseries_type == "regular":
            root_validator = dp_ts_root_validator_regular_wrapper(self.timeseries_params.time_step)
        elif self.timeseries_type == "irregular":
            root_validator = dp_ts_root_validator_irregular
        elif self.timeseries_type == "irregular_intervals":
            root_validator = dp_ts_root_validator_irregular_intervals
        else:
            raise ValueError(f"Unknown timeseries type '{self.timeseries_type}'")

        # Validators
        dp_validators = {
            "v_validator": field_validator("v")(dp_ts_v_validator),
            "root_validator": model_validator(mode="after")(root_validator),
        }

        self._dp_model = create_model(
            f"DataPointTimeseries_{self.id}",
            __base__=DataPointTimeseriesBase,
            __validators__=dp_validators,
            v=(create_model(f"DataPointTimeseriesValue_{self.id}", **dp_value_typing), ...),
        )

    @field_validator("series")
    @classmethod
    def add_default_series(cls, v, info: FieldValidationInfo):
        ts_type = info.data["timeseries_type"]
        default_series = timeseries_types[ts_type]["default_series"]

        for s in default_series:
            v[s] = TimeseriesSeries(**default_series[s])

        return v


"""A type union that covers AttrSpec class types:

- [AttrSpecPlain][dp3.common.attrspec.AttrSpecPlain]
- [AttrSpecObservations][dp3.common.attrspec.AttrSpecObservations]
- [AttrSpecTimeseries][dp3.common.attrspec.AttrSpecTimeseries]
"""
AttrSpecType = Union[AttrSpecTimeseries, AttrSpecObservations, AttrSpecPlain]


def AttrSpec(id: str, spec: dict[str, Any]) -> AttrSpecType:
    """Factory for `AttrSpec` classes"""

    assert isinstance(spec, dict), "Attribute specification must be a dict"
    if "type" not in spec:
        raise ValueError("Missing mandatory attribute `type`")
    attr_type = AttrType.from_str(spec.get("type"))
    subclasses = {
        AttrType.PLAIN: AttrSpecPlain,
        AttrType.OBSERVATIONS: AttrSpecObservations,
        AttrType.TIMESERIES: AttrSpecTimeseries,
    }
    return subclasses[attr_type](id=id, **spec)



"""
Functions to work with the main task queue (RabbitMQ)

There are two queues for each worker process:
- "normal" queue for tasks added by other components, this has a limit of 100
  tasks.
- "priority" one for tasks added by workers themselves, this has no limit since
  workers mustn't be stopped by waiting for the queue.

These queues are presented as a single one by this wrapper.
The TaskQueueReader first looks into the "priority" queue and only if there
is no task waiting, it reads the normal one.

Tasks are distributed to worker processes (and threads) by hash of the entity
which is to be modified. The destination queue is decided by the message source,
so each source must know how many worker processes are there.

Exchange and queues must be declared externally!

Related configuration keys and their defaults:
(should be part of global DP3 config files)
```
rabbitmq:
  host: localhost
  port: 5672
  virtual_host: /
  username: guest
  password: guest

worker_processes: 1
```
"""

import collections
import contextlib
import logging
import threading
import time
from typing import Callable, Union

import amqpstorm

from dp3.common.task import Task

# Exchange and queue names
# They must be pre-declared ('direct' exchange type) and binded.
# The names are prefixed by app_name
# Numbers from 0 to number_of_workers-1 are used as routing/binding keys.
DEFAULT_EXCHANGE = "{}-main-task-exchange"
DEFAULT_PRIORITY_EXCHANGE = "{}-priority-task-exchange"
DEFAULT_QUEUE = "{}-worker-{}"
DEFAULT_PRIORITY_QUEUE = "{}-worker-{}-pri"


# When reading, pre-fetch only a limited amount of messages
# (because pre-fetched messages are not counted to queue length limit)
PREFETCH_COUNT = 50

# number of seconds to wait for the i-th attempt to reconnect after error
RECONNECT_DELAYS = [1, 2, 5, 10, 30]


class QueueNotDeclared(RuntimeError):
    def __init__(self, queue_name):
        self.queue_name = queue_name

    def __str__(self):
        return (
            f"Queue '{self.queue_name}' is not declared in RabbitMQ! "
            "Run 'scripts/rmq_reconfigure.sh' to set up "
            "needed exchanges and queues on RabbitMQ server."
        )


class ExchangeNotDeclared(RuntimeError):
    def __init__(self, exchange_name):
        self.exchange_name = exchange_name

    def __str__(self):
        return (
            f"Exchange '{self.exchange_name}' is not declared in RabbitMQ! "
            f"Run 'scripts/rmq_reconfigure.sh' to set up "
            f"needed exchanges and queues on RabbitMQ server."
        )


class RobustAMQPConnection:
    """
    Common TaskQueue wrapper, handles connection to RabbitMQ server with automatic reconnection.
    TaskQueueWriter and TaskQueueReader are derived from this.

    Args:
        rabbit_config: RabbitMQ connection parameters, dict with following keys (all optional):
            host, port, virtual_host, username, password
    """

    def __init__(self, rabbit_config: dict = None) -> None:
        rabbit_config = {} if rabbit_config is None else rabbit_config
        self.log = logging.getLogger("RobustAMQPConnection")
        self.conn_params = {
            "hostname": rabbit_config.get("host", "localhost"),
            "port": int(rabbit_config.get("port", 5672)),
            "virtual_host": rabbit_config.get("virtual_host", "/"),
            "username": rabbit_config.get("username", "guest"),
            "password": rabbit_config.get("password", "guest"),
        }
        self.connection: amqpstorm.Connection = None
        self.channel: amqpstorm.Channel = None
        self._connection_id = 0

    def __del__(self):
        self.disconnect()

    def connect(self) -> None:
        """Create a connection (or reconnect after error).

        If connection can't be established, try it again indefinitely.
        """
        if self.connection:
            self.connection.close()
        self._connection_id += 1

        attempts = 0
        while True:
            attempts += 1
            try:
                self.connection = amqpstorm.Connection(**self.conn_params)
                self.log.debug(
                    "AMQP connection created, server: "
                    "'{hostname}:{port}/{virtual_host}'".format_map(self.conn_params)
                )
                if attempts > 1:
                    # This was a repeated attempt, print success message with ERROR level
                    self.log.error("... it's OK now, we're successfully connected!")

                self.channel = self.connection.channel()
                self.channel.confirm_deliveries()
                self.channel.basic.qos(PREFETCH_COUNT)
                break
            except amqpstorm.AMQPError as e:
                sleep_time = RECONNECT_DELAYS[min(attempts, len(RECONNECT_DELAYS)) - 1]
                self.log.error(
                    f"RabbitMQ connection error (will try to reconnect in {sleep_time}s): {e}"
                )
                time.sleep(sleep_time)
            except KeyboardInterrupt:
                break

    def disconnect(self) -> None:
        if self.connection:
            self.connection.close()
        self.connection = None
        self.channel = None

    def check_queue_existence(self, queue_name: str) -> bool:
        if queue_name is None:
            return True
        assert self.channel is not None, "not connected"
        try:
            self.channel.queue.declare(queue_name, passive=True)
        except amqpstorm.AMQPChannelError:
            return False
        return True

    def check_exchange_existence(self, exchange_name: str) -> bool:
        assert self.channel is not None, "not connected"
        try:
            self.channel.exchange.declare(exchange_name, passive=True)
        except amqpstorm.AMQPChannelError:
            return False
        return True


class TaskQueueWriter(RobustAMQPConnection):
    """
    Writes tasks into main Task Queue

    Args:
        app_name: DP3 application name (used as prefix for RMQ queues and exchanges)
        workers: Number of worker processes in the system
        rabbit_config: RabbitMQ connection parameters, dict with following keys (all optional):
            host, port, virtual_host, username, password
        exchange: Name of the exchange to write tasks to
            (default: `"<app-name>-main-task-exchange"`)
        priority_exchange: Name of the exchange to write priority tasks to
            (default: `"<app-name>-priority-task-exchange"`)
        parent_logger: Logger to inherit prefix from.
    """

    def __init__(
        self,
        app_name: str,
        workers: int = 1,
        rabbit_config: dict = None,
        exchange: str = None,
        priority_exchange: str = None,
        parent_logger: logging.Logger = None,
    ) -> None:
        rabbit_config = {} if rabbit_config is None else rabbit_config
        assert isinstance(workers, int) and workers >= 1, "count of workers must be positive number"
        assert isinstance(exchange, str) or exchange is None, "exchange argument has to be string!"
        assert (
            isinstance(priority_exchange, str) or priority_exchange is None
        ), "priority_exchange has to be string"

        super().__init__(rabbit_config)

        if parent_logger is not None:
            self.log = parent_logger.getChild("TaskQueueWriter")
        else:
            self.log = logging.getLogger("TaskQueueWriter")

        if exchange is None:
            exchange = DEFAULT_EXCHANGE.format(app_name)
        if priority_exchange is None:
            priority_exchange = DEFAULT_PRIORITY_EXCHANGE.format(app_name)

        self.workers = workers
        self.exchange = exchange
        self.exchange_pri = priority_exchange

    def check(self) -> bool:
        """
        Check that needed exchanges are declared, return True or raise RuntimeError.

        If needed exchanges are not declared, reconnect and try again. (max 5 times)
        """
        for attempt, sleep_time in enumerate(RECONNECT_DELAYS):
            if self.check_exchange_existence(self.exchange) and self.check_exchange_existence(
                self.exchange_pri
            ):
                return True
            self.log.warning(
                "RabbitMQ exchange configuration doesn't match (attempt %d of %d, retrying in %ds)",
                attempt + 1,
                len(RECONNECT_DELAYS),
                sleep_time,
            )
            time.sleep(sleep_time)
            self.disconnect()
            self.connect()
        if not self.check_exchange_existence(self.exchange):
            raise ExchangeNotDeclared(self.exchange)
        if not self.check_exchange_existence(self.exchange_pri):
            raise ExchangeNotDeclared(self.exchange_pri)
        return True

    def broadcast_task(self, task: Task, priority: bool = False) -> None:
        """
        Broadcast task to all workers

        Args:
            task: prepared task
            priority: if true, the task is placed into priority queue
                (should only be used internally by workers)
        """
        if not self.channel:
            self.connect()

        task_str = str(task) if len(str(task)) < 500 else f"{str(task)[:500]}...(truncated)"
        self.log.debug(f"Received new broadcast task: {task_str}")

        body = task.as_message()
        exchange = self.exchange_pri if priority else self.exchange

        for routing_key in range(self.workers):
            self._send_message(routing_key, exchange, body)

    def put_task(self, task: Task, priority: bool = False) -> None:
        """
        Put task (update_request) to the queue of corresponding worker

        Args:
            task: prepared task
            priority: if true, the task is placed into priority queue
                (should only be used internally by workers)
        """
        if not self.channel:
            self.connect()

        task_str = str(task) if len(str(task)) < 500 else f"{str(task)[:500]}...(truncated)"
        self.log.debug(f"Received new task: {task_str}")

        # Prepare routing key
        body = task.as_message()
        # index of the worker to send the task to
        routing_key = task.hashed_routing_key() % self.workers

        exchange = self.exchange_pri if priority else self.exchange
        self._send_message(routing_key, exchange, body)

    def _send_message(self, routing_key, exchange, body):
        """Send the message."""

        # ('mandatory' flag means that we want to guarantee it's delivered to someone.
        # If it can't be delivered (no consumer or full queue), wait a while and try again.
        # Always print just one error message for each unsuccessful message to be send.)
        self.log.debug(f"Sending a task with routing_key={routing_key} to exchange '{exchange}'")
        err_printed = 0
        while True:
            try:
                message = amqpstorm.Message.create(self.channel, body)
                success = message.publish(str(routing_key), exchange, mandatory=True)
                if success:  # message ACK'd
                    if err_printed == 1:
                        self.log.debug("... it's OK now, the message was successfully sent")
                    elif err_printed == 2:
                        self.log.warning("... it's OK now, the message was successfully sent")
                    else:
                        self.log.debug("Message successfully sent")
                    break
                else:  # message NACK'd
                    if err_printed != 1:
                        self.log.debug(
                            f"Message rejected (queue of worker {routing_key} is probably full), "
                            "will retry every 100ms"
                        )
                        err_printed = 1
                    time.sleep(0.1)
            except amqpstorm.AMQPChannelError as e:
                if err_printed != 2:
                    self.log.warning(
                        f"Can't deliver a message to worker {routing_key} "
                        f"(will retry every 5 seconds): {e}"
                    )
                    err_printed = 2
                time.sleep(5)
            except amqpstorm.AMQPConnectionError as e:
                self.log.error(f"RabbitMQ connection error (will try to reconnect): {e}")
                self.connect()


class TaskQueueReader(RobustAMQPConnection):
    """
    TaskQueueReader consumes messages from two RabbitMQ queues
    (normal and priority one for given worker)
    and passes them to the given callback function.

    Tasks from the priority queue are passed before the normal ones.

    Each received message must be acknowledged by calling `.ack(msg_tag)`.

    Args:
        callback: Function called when a message is received, prototype: func(tag, Task)
        parse_task: Function called to parse message body into a task, prototype: func(body) -> Task
        app_name: DP3 application name (used as prefix for RMQ queues and exchanges)
        worker_index: index of this worker
            (filled into DEFAULT_QUEUE string using .format() method)
        rabbit_config: RabbitMQ connection parameters, dict with following keys
            (all optional): host, port, virtual_host, username, password
        queue: Name of RabbitMQ queue to read from (default: `"<app-name>-worker-<index>"`)
        priority_queue: Name of RabbitMQ queue to read from (priority messages)
            or `False` to disable. (default: `"<app-name>-worker-<index>-pri"`)
        parent_logger: Logger to inherit prefix from.
    """

    def __init__(
        self,
        callback: Callable,
        parse_task: Callable[[str], Task],
        app_name: str,
        worker_index: int = 0,
        rabbit_config: dict = None,
        queue: str = None,
        priority_queue: Union[str, bool] = None,
        parent_logger: logging.Logger = None,
    ) -> None:
        rabbit_config = {} if rabbit_config is None else rabbit_config
        assert callable(callback), "callback must be callable object"
        assert (
            isinstance(worker_index, int) and worker_index >= 0
        ), "worker_index must be positive number"
        assert isinstance(queue, str) or queue is None, "queue must be string"
        assert (
            isinstance(priority_queue, str) or priority_queue is None or priority_queue is False
        ), "priority_queue must be string or False to disable"

        super().__init__(rabbit_config)

        if parent_logger is not None:
            self.log = parent_logger.getChild("TaskQueueReader")
        else:
            self.log = logging.getLogger("TaskQueueReader")

        self.callback = callback
        self.parse_task = parse_task

        if queue is None:
            queue = DEFAULT_QUEUE.format(app_name, worker_index)
        if priority_queue is None:
            priority_queue = DEFAULT_PRIORITY_QUEUE.format(app_name, worker_index)
        elif priority_queue is False:
            priority_queue = None
        self.queue_name = queue
        self.priority_queue_name = priority_queue
        self.worker_index = worker_index

        self.running = False

        self._consuming_thread = None
        self._processing_thread = None

        # Receive messages into 2 temporary queues
        # (max length should be equal to prefetch_count set in RabbitMQReader)
        self.cache = collections.deque()
        self.cache_pri = collections.deque()
        self.cache_full = threading.Event()  # signalize there's something in the cache

    def __del__(self):  # TODO is this needed?
        # self.log.debug("Destructor called")
        if hasattr(self, "_consuming_thread"):
            self._stop_consuming_thread()
        if hasattr(self, "_processing_thread"):
            self._stop_processing_thread()
        super().__del__()

    def start(self) -> None:
        """Start receiving tasks."""
        if self.running:
            raise RuntimeError("Already running")

        if not self.connection:
            self.connect()

        self.log.info("Starting TaskQueueReader")

        # Start thread for message consuming from server
        self.running = True
        self._consuming_thread = threading.Thread(target=self._consuming_thread_func)
        thread_n = self._consuming_thread.name.split("-")[-1]
        self._consuming_thread.name = f"Consumer-{self.worker_index}-{thread_n}"
        self._consuming_thread.start()

        # Start thread for message processing and passing to user's callback
        self._processing_thread = threading.Thread(target=self._msg_processing_thread_func)
        thread_n = self._processing_thread.name.split("-")[-1]
        self._processing_thread.name = f"Processor-{self.worker_index}-{thread_n}"
        self._processing_thread.start()

    def stop(self) -> None:
        """Stop receiving tasks."""
        if not self.running:
            raise RuntimeError("Not running")

        self.running = False
        self._stop_consuming_thread()
        self._stop_processing_thread()
        self.log.info("TaskQueueReader stopped")

    def reconnect(self) -> None:
        """Clear local message cache and reconnect to RabbitMQ server."""
        self.cache.clear()
        self.cache_pri.clear()

        self.connect()

    def check(self) -> bool:
        """
        Check that needed queues are declared, return True or raise RuntimeError.

        If needed queues are not declared, reconnect and try again. (max 5 times)
        """

        for attempt, sleep_time in enumerate(RECONNECT_DELAYS):
            if self.check_queue_existence(self.queue_name) and self.check_queue_existence(
                self.priority_queue_name
            ):
                return True
            self.log.warning(
                "RabbitMQ queue configuration doesn't match (attempt %d of %d, retrying in %ds)",
                attempt + 1,
                len(RECONNECT_DELAYS),
                sleep_time,
            )
            time.sleep(sleep_time)
            self.disconnect()
            self.connect()
        if not self.check_queue_existence(self.queue_name):
            raise QueueNotDeclared(self.queue_name)
        if not self.check_queue_existence(self.priority_queue_name):
            raise QueueNotDeclared(self.priority_queue_name)
        return True

    def ack(self, msg_tag: tuple[int, int]) -> bool:
        """Acknowledge processing of the message/task

        Will reconnect by itself on channel error.
        Args:
            msg_tag: Message tag received as the first param of the callback function.
        Returns:
            Whether the message was acknowledged successfully and can be processed further.
        """
        conn_id, msg_tag = msg_tag
        if conn_id != self._connection_id:
            return False
        try:
            self.channel.basic.ack(delivery_tag=msg_tag)
        except amqpstorm.AMQPChannelError as why:
            self.log.error("Channel error while acknowledging message: %s", why)
            self.reconnect()
            return False
        return True

    def _consuming_thread_func(self):
        # Register consumers and start consuming loop, reconnect on error
        while self.running:
            try:
                # Register consumers on both queues
                self.channel.basic.consume(self._on_message, self.queue_name, no_ack=False)
                if self.priority_queue_name is not None:
                    self.channel.basic.consume(
                        self._on_message_pri, self.priority_queue_name, no_ack=False
                    )
                # Start consuming (this call blocks until consuming is stopped)
                self.channel.start_consuming()
                self.channel.check_for_errors()
                self.log.info("Exiting thread - Channel was closed.")
                return
            except amqpstorm.AMQPChannelError as e:
                self.log.error("RabbitMQ channel error (will try to reconnect): %s", e)
                self.reconnect()
            except amqpstorm.AMQPConnectionError as e:
                self.log.error(f"RabbitMQ connection error (will try to reconnect): {e}")
                self.reconnect()

    # These two callbacks are called when a new message is received
    # - they only put the message into a local queue
    def _on_message(self, message):
        self.cache.append((self._connection_id, message))
        self.cache_full.set()

    def _on_message_pri(self, message):
        self.cache_pri.append((self._connection_id, message))
        self.cache_full.set()

    def _msg_processing_thread_func(self):
        # Reads local queues and passes tasks to the user callback.
        while self.running:
            # Get task from a local queue (try the priority one first)
            if len(self.cache_pri) > 0:
                conn_id, msg = self.cache_pri.popleft()
                pri = True
            elif len(self.cache) > 0:
                conn_id, msg = self.cache.popleft()
                pri = False
            else:
                self.cache_full.wait()
                self.cache_full.clear()
                continue

            body = msg.body
            tag = (conn_id, msg.delivery_tag)

            self.log.debug(
                "Received {}message: {} (tag: {})".format(
                    "priority " if pri else "",
                    body if len(body) < 500 else f"{body[:500]}...(truncated)",
                    tag,
                )
            )

            # Parse and check validity of received message
            try:
                task = self.parse_task(body)
            except (ValueError, TypeError, KeyError) as e:
                # Print error, acknowledge reception of the message and drop it
                self.log.error(
                    "Erroneous message received from main task queue. "
                    f"Error: {str(e)}, Message: '{body}'"
                )
                self.ack(tag)
                continue

            # Pass message to user's callback function
            try:
                self.callback(tag, task)
            except amqpstorm.AMQPChannelError as e:
                self.log.error("Channel error while processing message: %s", e)
                self.reconnect()
            except Exception as e:
                self.log.exception("Error in user callback function. %s: %s", type(e), str(e))
                self.log.error("Original message: %s", body)

    def watchdog(self):
        """
        Check whether both threads are running and perform a reset if not.

        Register to be called periodically by scheduler.
        """
        proc = self._processing_thread.is_alive()
        cons = self._consuming_thread.is_alive()

        if not proc or not cons:
            self.log.error(
                "Dead threads detected, processing=%s, consuming=%s, restarting TaskQueueReader.",
                "alive" if proc else "dead",
                "alive" if cons else "dead",
            )
            self._stop_consuming_thread()
            self._stop_processing_thread()

            self.channel.close()
            self.channel = None
            self.cache.clear()
            self.cache_pri.clear()

            self.connect()
            self.start()

    def _stop_consuming_thread(self) -> None:
        if self._consuming_thread:
            if self._consuming_thread.is_alive:
                # if not connected, no problem
                with contextlib.suppress(amqpstorm.AMQPError):
                    self.channel.stop_consuming()
            self._consuming_thread.join()
        self._consuming_thread = None

    def _stop_processing_thread(self) -> None:
        if self._processing_thread:
            if self._processing_thread.is_alive:
                self.running = False  # tell processing thread to stop
                self.cache_full.set()  # break potential wait() for data
            self._processing_thread.join()
        self._processing_thread = None


# Set up logging if not part of another program (i.e. when testing/debugging)
if __name__ == "__main__":
    # LOGFORMAT = "%(asctime)-15s,%(threadName)s,%(name)s,[%(levelname)s] %(message)s"
    LOGFORMAT = "[%(levelname)s] %(name)s: %(message)s"
    LOGDATEFORMAT = "%Y-%m-%dT%H:%M:%S"

    logging.basicConfig(level=logging.INFO, format=LOGFORMAT, datefmt=LOGDATEFORMAT)



import gzip
import json
import logging
import os
from datetime import datetime
from json import JSONEncoder
from pathlib import Path
from typing import Any, Optional

from pydantic import BaseModel, Extra

from dp3.common.attrspec import (
    AttrSpecObservations,
    AttrSpecType,
    AttrType,
    ObservationsHistoryParams,
)
from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import CronExpression, PlatformConfig
from dp3.common.types import ParsedTimedelta
from dp3.common.utils import entity_expired
from dp3.database.database import DatabaseError, EntityDatabase

DB_SEND_CHUNK = 100


class DatetimeEncoder(JSONEncoder):
    """JSONEncoder to encode datetime using the standard ADiCT format string."""

    def default(self, o: Any) -> Any:
        if isinstance(o, datetime):
            return o.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4]
        return super().default(o)


class SnapshotCleaningConfig(BaseModel):
    """Configuration for snapshot cleaning.

    Attributes:
        schedule: Schedule for snapshot cleaning.
        older_than: Snapshots older than this will be deleted.
    """

    schedule: CronExpression
    older_than: ParsedTimedelta


class DPArchivationConfig(BaseModel):
    """Configuration for datapoint archivation.

    Attributes:
        schedule: Schedule for datapoint archivation.
        older_than: Datapoints older than this will be archived.
        archive_dir: Directory where to archive datapoints. Can be `None` to only delete them.
    """

    schedule: CronExpression
    older_than: ParsedTimedelta
    archive_dir: Optional[str] = None


class HistoryManagerConfig(BaseModel, extra=Extra.forbid):
    """Configuration for history manager.

    Attributes:
        aggregation_schedule: Schedule for master document aggregation.
        datapoint_cleaning_schedule: Schedule for datapoint cleaning.
        mark_datapoints_schedule: Schedule for marking datapoints in master docs.
        snapshot_cleaning: Configuration for snapshot cleaning.
        datapoint_archivation: Configuration for datapoint archivation.
    """

    aggregation_schedule: CronExpression
    datapoint_cleaning_schedule: CronExpression
    mark_datapoints_schedule: CronExpression
    snapshot_cleaning: SnapshotCleaningConfig
    datapoint_archivation: DPArchivationConfig


class HistoryManager:
    def __init__(
        self, db: EntityDatabase, platform_config: PlatformConfig, registrar: CallbackRegistrar
    ) -> None:
        self.log = logging.getLogger("HistoryManager")

        self.db = db
        self.model_spec = platform_config.model_spec
        self.worker_index = platform_config.process_index
        self.num_workers = platform_config.num_processes
        self.config = HistoryManagerConfig.model_validate(
            platform_config.config.get("history_manager")
        )

        # Schedule master document aggregation
        registrar.scheduler_register(
            self.aggregate_master_docs, **self.config.aggregation_schedule.model_dump()
        )

        if platform_config.process_index != 0:
            self.log.debug(
                "History management will be disabled in this worker to avoid race conditions."
            )
            return

        # Schedule datapoints cleaning
        datapoint_marking_schedule = self.config.mark_datapoints_schedule
        registrar.scheduler_register(
            self.mark_datapoints_in_master_docs, **datapoint_marking_schedule.model_dump()
        )

        datapoint_cleaning_schedule = self.config.datapoint_cleaning_schedule
        registrar.scheduler_register(
            self.delete_old_dps, **datapoint_cleaning_schedule.model_dump()
        )

        snapshot_cleaning_schedule = self.config.snapshot_cleaning.schedule
        self.keep_snapshot_delta = self.config.snapshot_cleaning.older_than
        registrar.scheduler_register(
            self.delete_old_snapshots, **snapshot_cleaning_schedule.model_dump()
        )

        # Schedule datapoint archivation
        archive_config = self.config.datapoint_archivation
        self.keep_raw_delta = archive_config.older_than
        if archive_config.archive_dir is not None:
            self.log_dir = self._ensure_log_dir(archive_config.archive_dir)
        else:
            self.log_dir = None
        registrar.scheduler_register(self.archive_old_dps, **archive_config.schedule.model_dump())

    def delete_old_dps(self):
        """Deletes old data points from master collection."""
        self.log.debug("Deleting old records ...")

        for etype_attr, attr_conf in self.model_spec.attributes.items():
            etype, attr_name = etype_attr
            max_age = None

            if attr_conf.t == AttrType.OBSERVATIONS:
                max_age = attr_conf.history_params.max_age
            elif attr_conf.t == AttrType.TIMESERIES:
                max_age = attr_conf.timeseries_params.max_age

            if not max_age:
                continue

            t_old = datetime.utcnow() - max_age

            try:
                self.db.delete_old_dps(etype, attr_name, t_old)
            except DatabaseError as e:
                self.log.error(e)

    def mark_datapoints_in_master_docs(self):
        """Marks the timestamps of all datapoints in master documents."""
        self.log.debug("Marking the datapoint timestamps for all entity records ...")

        for entity, attr_conf in self.model_spec.entity_attributes.items():
            attrs_to_mark = []
            for attr, conf in attr_conf.items():
                if conf.t in AttrType.OBSERVATIONS | AttrType.TIMESERIES:
                    attrs_to_mark.append(attr)

            if not attrs_to_mark:
                continue
            try:
                res = self.db.mark_all_entity_dps_t2(entity, attrs_to_mark)
                self.log.debug("Marked %s records of %s", res.modified_count, entity)
            except DatabaseError as e:
                self.log.error(e)

    def delete_old_snapshots(self):
        """Deletes old snapshots."""
        t_old = datetime.now() - self.keep_snapshot_delta
        self.log.debug("Deleting all snapshots before %s", t_old)

        deleted_total = 0
        for etype in self.model_spec.entities:
            try:
                result = self.db.delete_old_snapshots(etype, t_old)
                deleted_total += result
            except DatabaseError as e:
                self.log.exception(e)
        self.log.debug("Deleted %s snapshots in total.", deleted_total)

    def archive_old_dps(self):
        """
        Archives old data points from raw collection.

        Updates already saved archive files, if present.
        """

        t_old = datetime.utcnow() - self.keep_raw_delta
        self.log.debug("Archiving all records before %s ...", t_old)

        for etype in self.model_spec.entities:
            res = self.db.move_raw_to_archive(etype)
            if res:
                self.log.info("Current %s raw collection was moved to archive: %s", etype, res)

        max_date, min_date, total_dps = self._get_raw_dps_summary(t_old)
        if total_dps == 0:
            self.log.debug("Found no datapoints to archive.")
            return
        self.log.debug(
            "Found %s datapoints to archive in the range %s - %s", total_dps, min_date, max_date
        )

        if self.log_dir is None:
            self.log.debug("No archive directory specified, skipping archivation.")
        else:
            min_date_string = min_date.strftime("%Y%m%dT%H%M%S")
            max_date_string = max_date.strftime("%Y%m%dT%H%M%S")
            date_logfile = self.log_dir / f"dp-log-{min_date_string}--{max_date_string}.jsonl"
            datapoints = 0

            with open(date_logfile, "w", encoding="utf-8") as logfile:
                for etype in self.model_spec.entities:
                    for result_cursor in self.db.get_archive(etype, after=min_date, before=t_old):
                        for dp in result_cursor:
                            logfile.write(
                                f"{json.dumps(self._reformat_dp(dp), cls=DatetimeEncoder)}\n"
                            )
                            datapoints += 1

            self.log.info("Archived %s datapoints to %s", datapoints, date_logfile)
            compress_file(date_logfile)
            os.remove(date_logfile)
            self.log.debug("Saved archive was compressed")

        deleted_count = 0
        for etype in self.model_spec.entities:
            for deleted_res in self.db.delete_old_archived_dps(etype, before=t_old):
                deleted_count += deleted_res.deleted_count
        self.log.info("Deleted %s datapoints", deleted_count)

        dropped_count = 0
        for etype in self.model_spec.entities:
            dropped_count += self.db.drop_empty_archives(etype)
        if dropped_count:
            self.log.info("Dropped %s empty archive collection(s)", dropped_count)

    @staticmethod
    def _reformat_dp(dp):
        del dp["_id"]
        dp["id"] = dp["eid"]
        dp["type"] = dp["etype"]
        del dp["eid"]
        del dp["etype"]
        return dp

    def _get_raw_dps_summary(
        self, before: datetime
    ) -> tuple[Optional[datetime], Optional[datetime], int]:
        date_ranges = []
        for etype in self.model_spec.entities:
            summary = self.db.get_archive_summary(etype, before=before)
            if summary:
                date_ranges.append(summary)
        if not date_ranges:
            return None, None, 0
        min_date = min(x["earliest"] for x in date_ranges)
        max_date = max(x["latest"] for x in date_ranges)
        total_dps = sum(x["count"] for x in date_ranges)
        return max_date, min_date, total_dps

    @staticmethod
    def _ensure_log_dir(log_dir_path: str):
        log_dir = Path(log_dir_path)
        log_dir.mkdir(parents=True, exist_ok=True)
        return log_dir

    def aggregate_master_docs(self):
        self.log.debug("Starting master documents aggregation.")
        start = datetime.now()
        utcnow = datetime.utcnow()
        entities = 0
        self.db.save_metadata(start, {"entities": 0, "aggregation_start": start})

        for entity in self.model_spec.entities:
            entity_attr_specs = self.model_spec.entity_attributes[entity]
            eids = []
            aggregated_records = []
            records_cursor = self.db.get_worker_master_records(
                self.worker_index, self.num_workers, entity, no_cursor_timeout=True
            )
            try:
                for master_document in records_cursor:
                    if entity_expired(utcnow, master_document):
                        continue  # Avoid expired entities to avoid conflict with garbage collector

                    entities += 1
                    self.aggregate_master_doc(entity_attr_specs, master_document)
                    eids.append(master_document["_id"])
                    aggregated_records.append(master_document)

                    if len(aggregated_records) > DB_SEND_CHUNK:
                        self.db.update_master_records(entity, eids, aggregated_records)
                        eids.clear()
                        aggregated_records.clear()

                if aggregated_records:
                    self.db.update_master_records(entity, eids, aggregated_records)
                    eids.clear()
                    aggregated_records.clear()
            finally:
                records_cursor.close()

        self.db.update_metadata(
            start, metadata={"aggregation_end": datetime.now()}, increase={"entities": entities}
        )
        self.log.debug("Master documents aggregation end.")

    @staticmethod
    def aggregate_master_doc(attr_specs: dict[str, AttrSpecType], master_document: dict):
        for attr, history in master_document.items():
            if attr not in attr_specs:
                continue
            spec = attr_specs[attr]

            if spec.t != AttrType.OBSERVATIONS or not spec.history_params.aggregate:
                continue

            if spec.multi_value:
                master_document[attr] = aggregate_multivalue_dp_history_on_equal(history, spec)
            else:
                master_document[attr] = aggregate_dp_history_on_equal(history, spec.history_params)


def aggregate_multivalue_dp_history_on_equal(history: list[dict], spec: AttrSpecObservations):
    """
    Merge multivalue datapoints in the history with equal values and overlapping time validity.

    Avergages the confidence.
    Will keep a pool of "active" datapoints and merge them with the next datapoint
    if they have the same value and overlapping time validity.

    FIXME:
      The average calculation only works for the current iteration,
      but for the next call of the algorithm, the count of aggregated datapoints is lost.
    """
    history = sorted(history, key=lambda x: x["t1"])
    aggregated_history = []
    pre = spec.history_params.pre_validity
    post = spec.history_params.post_validity

    if spec.data_type.hashable:
        current_dps = {}

        for dp in history:
            v = dp["v"]
            if v in current_dps:
                current_dp = current_dps[v]
                if current_dp["t2"] + post >= dp["t1"] - pre:  # Merge with current_dp
                    current_dp["t2"] = max(dp["t2"], current_dp["t2"])
                    current_dp["c"] += dp["c"]
                    current_dp["cnt"] += 1
                else:  # No overlap, finalize current_dp and reset
                    current_dp["c"] /= current_dp["cnt"]
                    del current_dp["cnt"]
                    aggregated_history.append(current_dp)
                    current_dps[v] = dp
                    current_dps[v]["cnt"] = 1
            else:  # New value, finalize initialize current_dp
                current_dps[v] = dp
                current_dps[v]["cnt"] = 1

        for _v, current_dp in current_dps.items():  # Finalize remaining dps
            current_dp["c"] /= current_dp["cnt"]
            del current_dp["cnt"]
            aggregated_history.append(current_dp)
        return aggregated_history
    else:
        current_dps = []

        for dp in history:
            v = dp["v"]
            for i, current_dp in enumerate(current_dps):
                if current_dp["v"] != v:
                    continue

                if current_dp["t2"] + post >= dp["t1"] - pre:  # Merge with current_dp
                    current_dp["t2"] = max(dp["t2"], current_dp["t2"])
                    current_dp["c"] += dp["c"]
                    current_dp["cnt"] += 1
                else:  # No overlap, finalize current_dp and reset
                    current_dp["c"] /= current_dp["cnt"]
                    del current_dp["cnt"]
                    aggregated_history.append(current_dp)
                    dp["cnt"] = 1
                    current_dps[i] = dp
                break
            else:  # New value, finalize initialize current_dp
                dp["cnt"] = 1
                current_dps.append(dp)

        for current_dp in current_dps:  # Finalize remaining dps
            current_dp["c"] /= current_dp["cnt"]
            del current_dp["cnt"]
            aggregated_history.append(current_dp)
        return aggregated_history


def aggregate_dp_history_on_equal(history: list[dict], spec: ObservationsHistoryParams):
    """
    Merge datapoints in the history with equal values and overlapping time validity.

    Avergages the confidence.

    FIXME:
      The average calculation only works for the current iteration,
      but for the next call of the algorithm, the count of aggregated datapoints is lost.
    """
    history = sorted(history, key=lambda x: x["t1"])
    aggregated_history = []
    current_dp = None
    merged_cnt = 0
    pre = spec.pre_validity
    post = spec.post_validity

    for dp in history:
        if not current_dp:
            current_dp = dp
            merged_cnt += 1
            continue

        if current_dp["v"] == dp["v"] and current_dp["t2"] + post >= dp["t1"] - pre:
            current_dp["t2"] = max(dp["t2"], current_dp["t2"])
            current_dp["c"] += dp["c"]
            merged_cnt += 1
        else:
            aggregated_history.append(current_dp)
            current_dp["c"] /= merged_cnt

            merged_cnt = 1
            current_dp = dp
    if current_dp:
        current_dp["c"] /= merged_cnt
        aggregated_history.append(current_dp)
    return aggregated_history


def compress_file(original: Path, compressed: Path = None):
    if compressed is None:
        compressed = original.parent / (original.name + ".gz")

    with (
        open(original, encoding="utf-8") as in_fp,
        gzip.open(compressed, "wt", encoding="utf-8") as out_fp,
    ):
        out_fp.writelines(in_fp)



import logging
import threading
import time
from datetime import datetime

from pymongo import ASCENDING, UpdateOne

from dp3.common.callback_registrar import CallbackRegistrar
from dp3.common.config import PlatformConfig
from dp3.common.datapoint import DataPointObservationsBase, DataPointTimeseriesBase
from dp3.common.task import DataPointTask
from dp3.database.database import EntityDatabase


class Telemetry:
    def __init__(
        self, db: EntityDatabase, platform_config: PlatformConfig, registrar: CallbackRegistrar
    ) -> None:
        self.log = logging.getLogger("Telemetry")

        self.db = db
        self.model_spec = platform_config.model_spec
        # self.config = platform_config.config.get("telemetry")  # No config for now
        self.cache_col = self.db.get_module_cache("Telemetry")

        self.local_cache = {}
        self.local_cache_lock = threading.Lock()

        # Schedule master document aggregation
        registrar.register_task_hook("on_task_start", self.note_latest_src_timestamp)
        mod = 30
        proc_i = platform_config.process_index
        n_proc = platform_config.num_processes
        spread_proc_index = proc_i * (mod // n_proc) if n_proc < mod else proc_i
        seconds = ",".join(f"{int(i)}" for i in range(60) if int(i - spread_proc_index) % mod == 0)
        registrar.scheduler_register(
            self.sync_to_db, second=seconds, minute="*", hour="*", misfire_grace_time=10
        )

    def note_latest_src_timestamp(self, task: DataPointTask):
        """Note the latest timestamp of each source in the task"""
        latest_timestamps = {}
        for dp in task.data_points:
            has_timestamp = isinstance(dp, (DataPointObservationsBase, DataPointTimeseriesBase))
            if dp.src is None or not has_timestamp:
                continue
            latest_timestamp = dp.t2 or dp.t1
            latest_timestamps[dp.src] = latest_timestamp

        if not latest_timestamps:
            return

        with self.local_cache_lock:
            self.local_cache.update(latest_timestamps)

    def sync_to_db(self):
        """Sync local timestamp cache to database."""
        with self.local_cache_lock:
            synced_cache = self.local_cache
            self.local_cache = {}

        updates = [
            UpdateOne(
                {"_id": src},
                [{"$set": {"src_t": {"$max": ["$src_t", latest_timestamp]}}}],
            )
            for src, latest_timestamp in synced_cache.items()
        ]

        if not updates:
            return

        try:
            start = time.time()
            res = self.cache_col.bulk_write(updates, ordered=False)
            end = time.time()
            self.log.debug(
                "Updating %s src_timestamp records: %s matched %s modified in %.4fs",
                len(updates),
                res.matched_count,
                res.modified_count,
                (end - start),
            )
            if len(updates) != res.matched_count:
                upserts = [
                    UpdateOne(
                        {"_id": src},
                        [{"$set": {"_id": src, "src_t": {"$max": ["$src_t", latest_timestamp]}}}],
                        upsert=True,
                    )
                    for src, latest_timestamp in synced_cache.items()
                ]
                start = time.time()
                res = self.cache_col.bulk_write(upserts, ordered=False)
                end = time.time()
                self.log.debug(
                    "Upserting %s src_timestamp records: %s matched %s modified in %.4fs",
                    len(upserts),
                    res.matched_count,
                    res.modified_count,
                    (end - start),
                )
        except Exception as e:
            self.log.error("Error updating src_timestamp records: %s", e)


class TelemetryReader:
    """Reader of telemetry data

    Used by API.
    Not contained inside `Telemetry` class due to usage of `CallbackRegistrar`
    and all of it's requirements (doesn't make sense for API).
    """

    def __init__(self, db: EntityDatabase) -> None:
        self.db = db
        self.cache_col = self.db.get_module_cache("Telemetry")

    def get_sources_validity(self) -> dict[str, datetime]:
        """Return timestamps (datetimes) of current validity of all sources."""
        src_data = self.cache_col.find({}).sort([("_id", ASCENDING)])

        return {src["_id"]: src["src_t"] for src in src_data}



from datetime import datetime

from fastapi import APIRouter

from dp3.api.internal.config import TELEMETRY_READER

router = APIRouter()


@router.get("/sources_validity")
async def get_sources_validity() -> dict[str, datetime]:
    """Get validity of all data sources

    Returns timestamps (datetimes) of current validity of all sources.
    This should be latest `t2` of incoming datapoints for given data source.
    """
    return TELEMETRY_READER.get_sources_validity()



import sys
from datetime import datetime, timedelta

import common

from dp3.api.internal.entity_response_models import EntityEidData, EntityEidMasterRecord

DATAPOINT_COUNT = 6


class GetEntityEidData(common.APITest):
    eid = None

    def setUp(self) -> None:
        super().setUpClass()
        t1 = datetime.now() - timedelta(minutes=30)
        t2 = t1 + timedelta(minutes=10)
        self.eid = f"test_get_data__{datetime.now()}"
        dp_base = {
            "src": "setup@test",
            "attr": "test_attr_history",
            "type": "test_entity_type",
            "id": self.eid,
        }
        dps = []
        for i in range(DATAPOINT_COUNT):
            dps.append(
                {
                    **dp_base,
                    "t1": t1.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4],
                    "t2": t2.strftime("%Y-%m-%dT%H:%M:%S.%fZ")[:-4],
                    "v": i,
                }
            )
            t1 += timedelta(minutes=10)
            t2 += timedelta(minutes=10)
        print(dps, file=sys.stderr)
        res = self.push_datapoints(dps)
        print(res.content.decode("utf-8"), file=sys.stderr)

    def test_get_entity_data(self):
        data = self.query_expected_value(
            lambda: self.get_entity_data(f"entity/test_entity_type/{self.eid}", EntityEidData),
            lambda data: "test_attr_history" in data.master_record,
        )
        self.assertEqual(DATAPOINT_COUNT, len(data.master_record["test_attr_history"]))

    def test_get_entity_master_record(self):
        data = self.query_expected_value(
            lambda: self.get_entity_data(
                f"entity/test_entity_type/{self.eid}/master", EntityEidMasterRecord
            ),
            lambda data: "test_attr_history" in data.root,
        )
        self.assertEqual(DATAPOINT_COUNT, len(data.root["test_attr_history"]))



from fastapi import APIRouter

from dp3.api.internal.config import CONTROL_WRITER, EnabledModules
from dp3.api.internal.response_models import SuccessResponse
from dp3.common.control import ControlAction, ControlMessage

router = APIRouter()


@router.get("/refresh_on_entity_creation")
async def refresh_on_entity_creation(etype: str) -> SuccessResponse:
    """Sends the action `refresh_on_entity_creation` into execution queue.

    This action is only accepted with the `etype` parameter.
    """
    CONTROL_WRITER.broadcast_task(
        ControlMessage(action=ControlAction.refresh_on_entity_creation, kwargs={"etype": etype})
    )
    return SuccessResponse(detail="Action sent.")


@router.get("/refresh_module_config")
async def refresh_module_config(module: EnabledModules) -> SuccessResponse:
    """Sends the action `refresh_module_config` into execution queue.

    This action is only accepted with the `module` parameter.
    """
    CONTROL_WRITER.broadcast_task(
        ControlMessage(action=ControlAction.refresh_module_config, kwargs={"module": module.value})
    )
    return SuccessResponse(detail="Action sent.")


@router.get("/{action}")
async def execute_action(action: ControlAction) -> SuccessResponse:
    """Sends the given action into execution queue."""
    CONTROL_WRITER.put_task(ControlMessage(action=action))
    return SuccessResponse(detail="Action sent.")



import logging
import threading
import time

from dp3.common.base_module import BaseModule


class MyModule(BaseModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._thread = None
        self._stop_event = threading.Event()
        self.log = logging.getLogger("MyModule")

    def start(self):
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()

    def stop(self):
        self._stop_event.set()
        self._thread.join()

    def _run(self):
        while not self._stop_event.is_set():
            self.log.info("Hello world!")
            time.sleep(1)



import common

from dp3.api.internal.entity_response_models import EntityState


class ListEntities(common.APITest):
    def test_get_entities(self):
        response = self.get_request("entities")
        self.assertEqual(200, response.status_code)

        received = response.json()
        self.assertIsInstance(received, dict)
        for key, item in received.items():
            received[key] = EntityState.model_validate(item)
        self.assertListEqual(list(received.keys()), list(common.MODEL_SPEC.entities.keys()))



"""
Module responsible for task
[distribution][dp3.task_processing.task_distributor],
[processing][dp3.task_processing.task_executor] and running configured
[hooks][dp3.task_processing.task_hooks]. Task distribution is possible due to the
[task queue][dp3.task_processing.task_queue].
"""



#!/usr/bin/env python3
"""
Run the DP3 API using uvicorn.
"""
import argparse
import sys

import uvicorn


def init_parser(parser):
    parser.add_argument(
        "--host",
        type=str,
        default="0.0.0.0",
        help="The host to bind to. (default: 0.0.0.0)",
    )
    parser.add_argument(
        "--port", type=int, default=5000, help="The port to bind to. (default: 5000)"
    )
    parser.add_argument(
        "--reload",
        action="store_true",
        default=False,
        help="Enable auto-reload of the api. (API changes will be picked up automatically.)",
    )


def run():
    print(
        "WARNING: The `api` entrypoint is deprecated due to possible namespace conflicts. "
        "Please use `dp3 api` instead.",
        file=sys.stderr,
    )

    parser = argparse.ArgumentParser(description="Run the DP3 API using uvicorn.")
    init_parser(parser)
    args = parser.parse_args()

    main(args)


def main(args):
    uvicorn.run("dp3.api.main:app", host=args.host, port=args.port, reload=args.reload)


if __name__ == "__main__":
    run()



#!/usr/bin/env python3
"""Simple script to migrate snapshots schema from flat single-snapshot documents to
a nested history-last schema.

Intended to be run only once to upgrade a database existing before 08-2024.
"""

import argparse
from collections import defaultdict

import bson
from pymongo import UpdateOne
from pymongo.errors import BulkWriteError, DocumentTooLarge, OperationFailure, WriteError

from dp3.common.config import ModelSpec, read_config_dir
from dp3.database.database import DatabaseError, EntityDatabase, MongoConfig

# Arguments parser
parser = argparse.ArgumentParser(
    description="Migrate snapshots schema from flat single-snapshot documents "
    "to a nested history schema."
)
parser.add_argument(
    "--config",
    default="/etc/adict/config",
    help="DP3 config directory (default: /etc/adict/config)",
)
parser.add_argument("--dry-run", action="store_true", help="Do not write to database")
parser.add_argument("-n", default=100, type=int, help="Number of updates to send per request")
args = parser.parse_args()

# Load DP3 configuration
config = read_config_dir(args.config, recursive=True)
model_spec = ModelSpec(config.get("db_entities"))

# Connect to database
connection_conf = MongoConfig.model_validate(config.get("database", {}))
client = EntityDatabase.connect(connection_conf)
print(client)
client.admin.command("ping")

db = client[connection_conf.db_name]

BSON_OBJECT_TOO_LARGE = 10334


def _migrate_to_oversized_snapshot(etype: str, eid: str, snapshot: dict):
    snapshot_col = f"{etype}#snapshots"
    os_col = f"{etype}#snapshots_oversized"

    if "_id" in snapshot:
        print(f"Removing _id {snapshot['_id']} from snapshot of {eid}")
        del snapshot["_id"]

    try:
        doc = db[snapshot_col].find_one_and_update(
            {"_id": eid},
            {
                "$set": {"oversized": True, "last": snapshot, "count": 0},
                "$unset": {"history": ""},
            },
        )
        inserts = list(doc.get("history", []))
        inserts.insert(0, doc.get("last", {}))
        inserts.insert(0, snapshot)
        db[os_col].insert_many(inserts)
    except Exception as e:
        raise DatabaseError(f"Update of snapshot {eid} failed: {e}, {snapshot}") from e


def save_snapshot(etype: str, snapshot: dict):
    """Saves snapshot to specified entity of current master document.

    Will move snapshot to oversized snapshots if the maintained bucket is too large.
    """
    if "eid" not in snapshot:
        return
    eid = snapshot["eid"]

    if "_id" in snapshot:
        print(f"Removing _id {snapshot['_id']} from snapshot of {eid}")
        del snapshot["_id"]

    snapshot_col = f"{etype}#snapshots"
    os_col = f"{etype}#snapshots_oversized"

    # Find out if the snapshot is oversized
    doc = db[snapshot_col].find_one({"_id": eid}, {"oversized": 1})
    if doc is None:
        # First snapshot of entity
        db[snapshot_col].insert_one(
            {"_id": eid, "last": snapshot, "history": [], "oversized": False, "count": 0}
        )
        print(f"Inserted snapshot of {eid}")
        return
    elif doc.get("oversized", False):
        # Snapshot is already marked as oversized
        db[snapshot_col].update_one({"_id": eid}, {"$set": {"last": snapshot}})
        db[os_col].insert_one(snapshot)
        return

    try:
        # Update a normal snapshot bucket
        res = db[snapshot_col].update_one(
            {"_id": eid},
            [
                {
                    "$set": {
                        "history": {
                            "$concatArrays": [
                                ["$last"],
                                "$history",
                            ]
                        },
                        "count": {"$sum": ["$count", 1]},
                    }
                },
                {"$set": {"last": {"$literal": snapshot}}},
            ],
        )
        if res.modified_count == 0:
            print(f"Snapshot of {eid} was not updated, {res.raw_result}")
    except (WriteError, OperationFailure) as e:
        if e.code != BSON_OBJECT_TOO_LARGE:
            raise e
        # The snapshot is too large, move it to oversized snapshots
        print(f"Snapshot of {eid} is too large: {e}, marking as oversized.")
        _migrate_to_oversized_snapshot(etype, eid, snapshot)
    except Exception as e:
        raise DatabaseError(f"Insert of snapshot {eid} failed: {e}, {snapshot}") from e


def save_snapshots(etype: str, snapshots: list[dict]):
    """
    Saves a list of snapshots of current master documents.

    All snapshots must belong to same entity type.

    Will move snapshots to oversized snapshots if the maintained bucket is too large.
    For better understanding, see `save_snapshot()`.

    """
    snapshots_by_eid = defaultdict(list)
    for snapshot in snapshots:
        if "eid" not in snapshot:
            continue
        if "_id" in snapshot:
            del snapshot["_id"]
        snapshots_by_eid[snapshot["eid"]].append(snapshot)
    print(f"Saving {len(snapshots)} snapshots of {len(snapshots_by_eid)} entities of {etype}")

    snapshot_col = f"{etype}#snapshots"
    os_col = f"{etype}#snapshots_oversized"

    # Find out if any of the snapshots are oversized
    docs = list(
        db[snapshot_col].find(
            {"_id": {"$in": list(snapshots_by_eid.keys())}}, {"oversized": 1, "eid": 1}
        )
    )

    updates = []
    update_originals = []
    oversized_inserts = []
    oversized_updates = []

    for doc in docs:
        eid = doc["_id"]
        if not doc.get("oversized", False):
            # A normal snapshot, shift the last snapshot to history and update last
            updates.append(
                UpdateOne(
                    {"_id": eid},
                    [
                        {
                            "$set": {
                                "history": {
                                    "$concatArrays": [
                                        snapshots_by_eid[eid][:-1],
                                        ["$last"],
                                        "$history",
                                    ]
                                },
                                "count": {"$sum": [len(snapshots_by_eid[eid]), "$count"]},
                            }
                        },
                        {"$set": {"last": {"$literal": snapshots_by_eid[eid][-1]}}},
                    ],
                )
            )
            update_originals.append(snapshots_by_eid[eid])
        else:
            # Snapshot is already marked as oversized
            oversized_inserts.extend(snapshots_by_eid[eid])
            oversized_updates.append(
                UpdateOne({"_id": eid}, {"$set": {"last": snapshots_by_eid[eid][-1]}})
            )
        del snapshots_by_eid[eid]

    # The remaining snapshots are new
    inserts = [
        {
            "_id": eid,
            "last": eid_snapshots[-1],
            "history": eid_snapshots[:-1],
            "oversized": False,
            "count": len(eid_snapshots) - 1,
        }
        for eid, eid_snapshots in snapshots_by_eid.items()
    ]

    if updates:
        try:
            res = db[snapshot_col].bulk_write(updates, ordered=False)
            if res.modified_count != len(updates):
                print(
                    f"Some snapshots were not updated, "
                    f"{res.modified_count} != {len(snapshots_by_eid)}"
                )
        except (BulkWriteError, OperationFailure) as e:
            print("Update of snapshots failed, will retry with oversize.")
            failed_indexes = [
                err["index"]
                for err in e.details["writeErrors"]
                if err["code"] == BSON_OBJECT_TOO_LARGE
            ]
            failed_snapshots = (update_originals[i] for i in failed_indexes)
            for eid_snapshots in failed_snapshots:
                eid = eid_snapshots[0]["eid"]
                failed_snapshots = sorted(
                    eid_snapshots, key=lambda s: s["_time_created"], reverse=True
                )
                _migrate_to_oversized_snapshot(etype, eid, failed_snapshots[0])
                oversized_inserts.extend(failed_snapshots[1:])

            if any(err["code"] != BSON_OBJECT_TOO_LARGE for err in e.details["writeErrors"]):
                # Some other error occurred
                raise e
        except Exception as e:
            raise DatabaseError(f"Update of snapshots failed: {str(e)[:2048]}") from e

    if inserts:
        try:
            # Insert new snapshots
            res = db[snapshot_col].insert_many(inserts, ordered=False)
            if len(res.inserted_ids) != len(snapshots_by_eid):
                print(
                    f"Some snapshots were not inserted, "
                    f"{len(res.inserted_ids)} != {len(snapshots_by_eid)}"
                )
        except (DocumentTooLarge, OperationFailure) as e:
            print(f"Snapshot too large: {e}")
            checked_inserts = []
            oversized_inserts = []

            # Filter out the oversized snapshots
            for insert_doc in inserts:
                bsize = len(bson.BSON.encode(insert_doc))
                if bsize < 16 * 1024 * 1024:
                    checked_inserts.append(insert_doc)
                else:
                    eid = insert_doc["_id"]
                    checked_inserts.append(
                        {
                            "_id": eid,
                            "last": insert_doc["last"],
                            "oversized": True,
                            "history": [],
                            "count": 0,
                        }
                    )
                    oversized_inserts.extend(insert_doc["history"] + [insert_doc["last"]])
            try:
                db[snapshot_col].insert_many(checked_inserts, ordered=False)
            except Exception as e:
                raise DatabaseError(f"Insert of snapshots failed: {e}") from e
        except Exception as e:
            raise DatabaseError(f"Insert of snapshot failed: {e}") from e

    # Update the oversized snapshots
    if oversized_inserts:
        try:
            if oversized_updates:
                db[snapshot_col].bulk_write(oversized_updates)
            db[os_col].insert_many(oversized_inserts)
        except Exception as e:
            raise DatabaseError(f"Insert of snapshots failed: {str(e)[:2048]}") from e


for entity in model_spec.entities:
    if len(list(db.list_collections(filter={"name": f"{entity}#snapshots_old"}))) == 0:
        db[f"{entity}#snapshots"].rename(f"{entity}#snapshots_old")
    print(entity)

    snapshots = []
    for record in db[f"{entity}#snapshots_old"].find({}, sort=[("_time_created", 1)]):
        del record["_id"]
        snapshots.append(record)
        if len(snapshots) >= args.n:
            if not args.dry_run:
                save_snapshots(entity, snapshots)
            else:
                print(f"Would save {len(snapshots)} snapshots")
            snapshots.clear()
    if snapshots and not args.dry_run:
        save_snapshots(entity, snapshots)
    elif snapshots:
        print(f"Would save {len(snapshots)} snapshots")



"""
Update the database schema after making conflicting changes to the model.

Authors: Ondřej Sedláček <xsedla1o@stud.fit.vutbr.cz>
"""

import logging

from dp3.common.config import ModelSpec, read_config_dir
from dp3.database.database import EntityDatabase


def init_parser(parser):
    parser.add_argument(
        "config",
        metavar="CONFIG_DIR",
        help="Path to a directory containing configuration files (e.g. /etc/my_app/config)",
    )
    parser.add_argument(
        "--bypass",
        action="store_true",
        default=False,
        help="Bypass the suggested database changes and update the schema regardless.",
    )


def confirm_changes(prompt: str):
    while True:
        answer = input(prompt).lower()
        if answer == "" or answer[0] == "n":
            return False
        if answer[0] == "y":
            return True


def main(args):
    # Load DP3 configuration
    config = read_config_dir(args.config, recursive=True)

    # Setup logging
    LOGFORMAT = "%(asctime)-15s,%(name)s,[%(levelname)s] %(message)s"
    LOGDATEFORMAT = "%Y-%m-%dT%H:%M:%S"
    logging.basicConfig(level=logging.DEBUG, format=LOGFORMAT, datefmt=LOGDATEFORMAT)
    log = logging.getLogger("SchemaUpdate")

    # Connect to database
    db = EntityDatabase(
        config,
        ModelSpec(config.get("db_entities")),
        config.get("processing_core.worker_processes"),
    )

    prev_schema, config_schema, updates, deleted_entites = db.schema_cleaner.get_schema_status()
    if prev_schema["version"] != config_schema["version"]:
        log.info(
            f"Schema version mismatch: {prev_schema['version']} (DB) "
            f"!= {config_schema['version']} (config)"
        )
        if confirm_changes("Are you sure you want to perform a migration now? (y/[n]): "):
            db.schema_cleaner.migrate(prev_schema)
        return

    if prev_schema["storage"] != config_schema["storage"]:
        log.info(
            f"Storage mismatch: {prev_schema['storage']} (DB) "
            f"!= {config_schema['storage']} (config)"
        )
        if confirm_changes("Are you sure you want to change the storage now? (y/[n]): "):
            db.schema_cleaner.update_storage(prev_schema["storage"], config_schema["storage"])
    elif prev_schema["schema"] == config_schema["schema"]:
        log.info("Schema is OK!")
        return

    if not updates and not deleted_entites:
        db.schema_cleaner.schemas.insert_one(config_schema)
        log.info("Updated schema without any changes to master records, OK now!")
        return

    if deleted_entites:
        log.info("Suggested removal of entities:")
        for entity in deleted_entites:
            log.info(f"- {entity}")

    if updates:
        log.info("Suggested changes to master records:")
        for entity, entity_updates in updates.items():
            log.info(f"- {entity}: {dict(entity_updates)}")

    if args.bypass:
        if not confirm_changes(
            "Are you sure you want update the schema without applying changes? (y/[n]): "
        ):
            log.info("Aborted schema update.")
            return
        db.schema_cleaner.schemas.insert_one(config_schema)
        log.info("Bypassing suggested changes, updated schema, OK now!")
        return

    if not confirm_changes("Are you sure you want to apply these changes? (y/[n]): "):
        log.info("Aborted schema update.")
        return
    db.schema_cleaner.execute_updates(updates, deleted_entites)
    db.schema_cleaner.schemas.insert_one(config_schema)
    log.info("Applied suggested changes, updated schema, OK now!")



"""
Common modules which are used throughout the platform.

- [Config][dp3.common.config], [EntitySpec][dp3.common.entityspec] and
[AttrSpec][dp3.common.attrspec] - Models for reading, validation and representing
platform configuration of entities and their attributes.
[datatype][dp3.common.datatype] is also used within this context.
- [Scheduler][dp3.common.scheduler] - Allows modules to run callbacks at specified times
- [Task][dp3.common.task.Task] - Model for a single task processed by the platform
- [Utils][dp3.common.utils] - Auxiliary utility functions
"""



from datetime import datetime
from typing import Any, Optional

from fastapi import APIRouter, Depends, HTTPException, Query, Request
from pydantic import Json, NonNegativeInt, ValidationError

from dp3.api.internal.config import DB, MODEL_SPEC, TASK_WRITER
from dp3.api.internal.entity_response_models import (
    EntityEidAttrValue,
    EntityEidAttrValueOrHistory,
    EntityEidData,
    EntityEidList,
    EntityEidMasterRecord,
    EntityEidSnapshots,
)
from dp3.api.internal.helpers import api_to_dp3_datapoint
from dp3.api.internal.models import (
    DataPoint,
)
from dp3.api.internal.response_models import ErrorResponse, RequestValidationError, SuccessResponse
from dp3.common.attrspec import AttrType
from dp3.common.task import DataPointTask, task_context
from dp3.database.database import DatabaseError


async def check_etype(etype: str):
    """Middleware to check entity type existence"""
    if etype not in MODEL_SPEC.entities:
        raise RequestValidationError(["path", "etype"], f"Entity type '{etype}' doesn't exist")
    return etype


def get_eid_master_record_handler(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
):
    """Handler for getting master record of EID"""
    # TODO: This is probably not the most efficient way. Maybe gather only
    # plain data from master record and then call `get_timeseries_history`
    # for timeseries.
    master_record = DB.get_master_record(etype, eid)
    if "_id" in master_record:
        del master_record["_id"]
    if "#hash" in master_record:
        del master_record["#hash"]

    entity_attribs = MODEL_SPEC.attribs(etype)

    # Get filtered timeseries data
    for attr in master_record:
        # Check for no longer existing attributes
        if attr in entity_attribs and entity_attribs[attr].t == AttrType.TIMESERIES:
            master_record[attr] = DB.get_timeseries_history(
                etype, attr, eid, t1=date_from, t2=date_to
            )

    return master_record


def get_eid_snapshots_handler(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
):
    """Handler for getting snapshots of EID"""
    snapshots = list(DB.get_snapshots(etype, eid, t1=date_from, t2=date_to))

    return snapshots


router = APIRouter(dependencies=[Depends(check_etype)])

# As variable, because otherwise generates ruff B008 error
eid_filter_query_param = Query(default="", deprecated=True)


@router.get(
    "/{etype}", responses={400: {"description": "Query can't be processed", "model": ErrorResponse}}
)
async def list_entity_type_eids(
    etype: str,
    eid_filter: str = eid_filter_query_param,
    fulltext_filters: Json = None,
    generic_filter: Json = None,
    skip: NonNegativeInt = 0,
    limit: NonNegativeInt = 20,
) -> EntityEidList:
    """List latest snapshots of all `id`s present in database under `etype`.

    Contains only latest snapshot.

    Uses pagination.
    Setting `limit` to 0 is interpreted as no limit (return all results).

    Returns only documents matching `generic_filter` and `fulltext_filters`
    (JSON object in format: attribute - fulltext filter).
    Fulltext filters are interpreted as regular expressions.
    Only string values may be filtered this way. There's no validation that queried attribute
    can be fulltext filtered.
    Only plain and observation attributes with string-based data types can be queried.
    Array and set data types are supported as well as long as they are not multi value
    at the same time.
    If you need to filter EIDs, use attribute `eid` (`eid_filter` is deprecated and you should
    migrate to `fulltext_filters["eid"]`).

    Generic filter allows filtering using generic MongoDB query (including `$and`, `$or`,
    `$lt`, etc.).
    There are no attribute name checks (may be added in the future).

    Generic and fulltext filters are merged - fulltext overrides conflicting keys.
    """
    if not fulltext_filters:
        fulltext_filters = {}
    if not isinstance(fulltext_filters, dict):
        raise HTTPException(status_code=400, detail="Fulltext filter is invalid")

    if not generic_filter:
        generic_filter = {}
    if not isinstance(generic_filter, dict):
        raise HTTPException(status_code=400, detail="Generic filter is invalid")

    for attr in fulltext_filters:
        ftr = fulltext_filters[attr]
        if not isinstance(ftr, str):
            raise HTTPException(status_code=400, detail=f"Filter '{ftr}' is not string")

    # `eid_filter` is deprecated - to be removed in the future
    if eid_filter:
        fulltext_filters["eid"] = eid_filter

    try:
        cursor, total_count = DB.get_latest_snapshots(etype, fulltext_filters, generic_filter)
        cursor_page = cursor.skip(skip).limit(limit)
    except DatabaseError as e:
        raise HTTPException(status_code=400, detail="Query is invalid") from e

    time_created = None

    # Remove _id field
    result = [r["last"] for r in cursor_page]
    for r in result:
        time_created = r["_time_created"]
        del r["_time_created"]

    return EntityEidList(
        time_created=time_created, count=len(result), total_count=total_count, data=result
    )


@router.get("/{etype}/{eid}")
async def get_eid_data(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
) -> EntityEidData:
    """Get data of `etype`'s `eid`.

    Contains all snapshots and master record.
    Snapshots are ordered by ascending creation time.

    Combines function of `/{etype}/{eid}/master` and `/{etype}/{eid}/snapshots`.
    """
    master_record = get_eid_master_record_handler(etype, eid, date_from, date_to)
    snapshots = get_eid_snapshots_handler(etype, eid, date_from, date_to)

    # Whether this eid contains any data
    empty = not master_record and len(snapshots) == 0

    return EntityEidData(empty=empty, master_record=master_record, snapshots=snapshots)


@router.get("/{etype}/{eid}/master")
async def get_eid_master_record(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
) -> EntityEidMasterRecord:
    """Get master record of `etype`'s `eid`."""
    return get_eid_master_record_handler(etype, eid, date_from, date_to)


@router.get("/{etype}/{eid}/snapshots")
async def get_eid_snapshots(
    etype: str, eid: str, date_from: Optional[datetime] = None, date_to: Optional[datetime] = None
) -> EntityEidSnapshots:
    """Get snapshots of `etype`'s `eid`."""
    return get_eid_snapshots_handler(etype, eid, date_from, date_to)


@router.get("/{etype}/{eid}/get/{attr}")
async def get_eid_attr_value(
    etype: str,
    eid: str,
    attr: str,
    date_from: Optional[datetime] = None,
    date_to: Optional[datetime] = None,
) -> EntityEidAttrValueOrHistory:
    """Get attribute value

    Value is either of:
    - current value: in case of plain attribute
    - current value and history: in case of observation attribute
    - history: in case of timeseries attribute
    """
    # Check if attribute exists
    if attr not in MODEL_SPEC.attribs(etype):
        raise RequestValidationError(["path", "attr"], f"Attribute '{attr}' doesn't exist")

    value_or_history = DB.get_value_or_history(etype, attr, eid, t1=date_from, t2=date_to)

    return EntityEidAttrValueOrHistory(attr_type=MODEL_SPEC.attr(etype, attr).t, **value_or_history)


@router.post("/{etype}/{eid}/set/{attr}")
async def set_eid_attr_value(
    etype: str, eid: str, attr: str, body: EntityEidAttrValue, request: Request
) -> SuccessResponse:
    """Set current value of attribute

    Internally just creates datapoint for specified attribute and value.

    This endpoint is meant for `editable` plain attributes -- for direct user edit on DP3 web UI.
    """
    # Check if attribute exists
    if attr not in MODEL_SPEC.attribs(etype):
        raise RequestValidationError(["path", "attr"], f"Attribute '{attr}' doesn't exist")

    # Construct datapoint
    try:
        dp = DataPoint(
            type=etype,
            id=eid,
            attr=attr,
            v=body.value,
            t1=datetime.now(),
            src=f"{request.client.host} via API",
        )
        dp3_dp = api_to_dp3_datapoint(dp.dict())
    except ValidationError as e:
        raise RequestValidationError(["body", "value"], e.errors()[0]["msg"]) from e

    # This shouldn't fail
    with task_context(MODEL_SPEC):
        task = DataPointTask(etype=etype, eid=eid, data_points=[dp3_dp])

    # Push tasks to task queue
    TASK_WRITER.put_task(task, False)

    # Datapoints from this endpoint are intentionally not logged using `DPLogger`.
    # If for some reason, in the future, they need to be, just copy code from data ingestion
    # endpoint.

    return SuccessResponse()


@router.get(
    "/{etype}/_/distinct/{attr}",
    responses={400: {"description": "Query can't be processed", "model": ErrorResponse}},
)
async def get_distinct_attribute_values(etype: str, attr: str) -> dict[Any, int]:
    """Gets distinct attribute values and their counts based on latest snapshots

    Useful for displaying `<select>` enumeration fields.

    Works for all plain and observation data types except `dict` and `json`.
    """
    try:
        return DB.get_distinct_val_count(etype, attr)
    except DatabaseError as e:
        raise HTTPException(status_code=400, detail=str(e)) from e


@router.post("/{etype}/{eid}/ttl")
async def extend_eid_ttls(etype: str, eid: str, body: dict[str, datetime]) -> SuccessResponse:
    """Extend TTLs of the specified entity"""
    # Construct task
    with task_context(MODEL_SPEC):
        task = DataPointTask(etype=etype, eid=eid, ttl_tokens=body)

    # Push tasks to task queue
    TASK_WRITER.put_task(task, False)

    return SuccessResponse()


@router.delete("/{etype}/{eid}")
async def delete_eid_record(etype: str, eid: str) -> SuccessResponse:
    """Delete the master record and snapshots of the specified entity.

    Notice that this does not delete any raw datapoints,
    or block the re-creation of the entity if new datapoints are received.
    """
    # Create a "delete" task and push it to task queue
    with task_context(MODEL_SPEC):
        task = DataPointTask(etype=etype, eid=eid, delete=True)
    TASK_WRITER.put_task(task, False)

    return SuccessResponse()



# Very simple datapoint generator for bus example config

import datetime
import json
import random


class TimeContainer:
    def __init__(self):
        self.time = datetime.datetime.utcnow() - datetime.timedelta(days=4)

    def add_minutes(self, minutes: int):
        self.time += datetime.timedelta(minutes=minutes)
        return self.time

    def add_minutes_no_modify(self, minutes: int):
        return self.time + datetime.timedelta(minutes=minutes)


time = TimeContainer()


def random_initial_location():
    latitude = random.uniform(39.0, 41.0)
    longitute = random.uniform(18.0, 22.0)

    return [round(latitude, 3), round(longitute, 3)]


def do_random_location_increment(current_location):
    current_location[0] = round(current_location[0] + random.uniform(-1.0, 1.0), 3)
    current_location[1] = round(current_location[1] + random.uniform(-1.0, 1.0), 3)


def random_t1(pseudo_current_time):
    return pseudo_current_time.add_minutes(random.randint(0, 60))


def random_t2(pseudo_current_time):
    return pseudo_current_time.add_minutes_no_modify(random.randint(5, 15))


def random_passenger_counts_3():
    return [random.randint(0, 40), random.randint(0, 40), random.randint(0, 40)]


dps = []

bus_lines = {
    "10": {"label": "Special bus 10", "location": random_initial_location()},
    "11": {"label": "Special bus 11", "location": random_initial_location()},
    "12": {"label": "Bus 12", "location": random_initial_location()},
    "19": {"label": "Bus 19 ", "location": random_initial_location()},
    "48": {"label": "Bus 48", "location": random_initial_location()},
    "93": {"label": "Bus 93", "location": random_initial_location()},
    "E73": {"label": "Express bus 73", "location": random_initial_location()},
    "N72": {"label": "Night bus 72", "location": random_initial_location()},
    "S20": {"label": "School bus 20", "location": random_initial_location()},
}

for n in bus_lines:
    bus_line = bus_lines[n]

    dps.append(
        {
            "type": "bus",
            "id": n,
            "attr": "label",
            "v": bus_line["label"],
            "t1": random_t1(
                time,
            ).isoformat(),
            "t2": random_t2(
                time,
            ).isoformat(),
            "src": "Manual",
        }
    )

    for _i in range(5):
        dps.append(
            {
                "type": "bus",
                "id": n,
                "attr": "location",
                "v": bus_line["location"],
                "t1": random_t1(
                    time,
                ).isoformat(),
                "t2": random_t2(
                    time,
                ).isoformat(),
                "src": "Tracking SW",
            }
        )
        do_random_location_increment(bus_line["location"])

    for _i in range(5):
        dps.append(
            {
                "type": "bus",
                "id": n,
                "attr": "speed",
                "v": round(random.uniform(0.0, 100.0), 1),
                "t1": random_t1(
                    time,
                ).isoformat(),
                "t2": random_t2(
                    time,
                ).isoformat(),
                "src": "Tracking SW",
            }
        )

    for _i in range(5):
        random_t1_local = random_t1(
            time,
        )

        dps.append(
            {
                "type": "bus",
                "id": n,
                "attr": "passengers_in_out",
                "v": {
                    "front_in": random_passenger_counts_3(),
                    "front_out": random_passenger_counts_3(),
                    "middle_in": random_passenger_counts_3(),
                    "middle_out": random_passenger_counts_3(),
                    "back_in": random_passenger_counts_3(),
                    "back_out": random_passenger_counts_3(),
                },
                "t1": random_t1_local.isoformat(),
                "t2": (random_t1_local + datetime.timedelta(minutes=30)).isoformat(),
                "src": "Bus counter",
            }
        )

print(json.dumps(dps, indent=2))



import hashlib
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Optional

from pydantic import (
    AfterValidator,
    BaseModel,
    BeforeValidator,
    ValidationError,
    field_validator,
)
from pydantic_core.core_schema import FieldValidationInfo

from dp3.common.config import ModelSpec
from dp3.common.datapoint import DataPointBase

_init_context_var = ContextVar("_init_context_var", default=None)


def HASH(key: str) -> int:
    """Hash function used to distribute tasks to worker processes.
    Args:
        key: to be hashed
    Returns:
        last 4 bytes of MD5
    """
    return int(hashlib.md5(key.encode("utf8")).hexdigest()[-4:], 16)


@contextmanager
def task_context(model_spec: ModelSpec) -> Iterator[None]:
    """Context manager for setting the `model_spec` context variable."""
    token = _init_context_var.set({"model_spec": model_spec})
    try:
        yield
    finally:
        _init_context_var.reset(token)


class Task(BaseModel, ABC):
    """
    A generic task type class.

    An abstraction for the [task queue][dp3.task_processing.task_queue] classes to depend upon.
    """

    @abstractmethod
    def routing_key(self) -> str:
        """
        Returns:
            A string to be used as a routing key between workers.
        """

    def hashed_routing_key(self) -> int:
        """
        Returns:
            An integer to be used as a hashed routing key between workers.
        """
        return HASH(self.routing_key())

    @abstractmethod
    def as_message(self) -> str:
        """
        Returns:
            A string representation of the object.
        """


def instanciate_dps(v, info: FieldValidationInfo):
    # If already instantiated, return
    if isinstance(v, DataPointBase):
        return v

    etype = v.get("etype")
    attr = v.get("attr")

    # Fetch datapoint model
    context = info.context
    assert context.get("model_spec"), "Missing `model_spec` in context"
    try:
        dp_model = context.get("model_spec").attr(etype, attr).dp_model
    except (KeyError, TypeError) as e:
        raise ValueError(f"Attribute '{attr}' not found in entity '{etype}'") from e

    # Parse datapoint using model from attribute specification
    try:
        return dp_model.model_validate(v)
    except ValidationError as e:
        raise ValueError(e) from e


def validate_data_points(v, info: FieldValidationInfo):
    assert (
        v.etype == info.data["etype"]
    ), f"Task's etype '{info.data['etype']}' doesn't match datapoint's etype '{v.etype}'"
    assert (
        v.eid == info.data["eid"]
    ), f"Task's eid '{info.data['eid']}' doesn't match datapoint's eid '{v.eid}'"
    return v


ValidatedDataPoint = Annotated[
    DataPointBase,
    BeforeValidator(instanciate_dps),
    AfterValidator(validate_data_points),
]


class DataPointTask(Task):
    """DataPointTask

    Contains single task to be pushed to TaskQueue and processed.
    Attributes:
        etype: Entity type
        eid: Entity id / key
        data_points: List of DataPoints to process
        tags: List of tags
        ttl_tokens: Dictionary of TTL tokens.
        delete: If True, delete entity
    """

    etype: str
    eid: str
    data_points: list[ValidatedDataPoint] = []
    tags: list[Any] = []
    ttl_tokens: Optional[dict[str, datetime]] = None
    delete: bool = False

    def __init__(__pydantic_self__, **data: Any) -> None:
        """Set the `model_spec` context variable and initialize the Task.

        See https://docs.pydantic.dev/latest/concepts/validators/#validation-context
        """
        __pydantic_self__.__pydantic_validator__.validate_python(
            data,
            self_instance=__pydantic_self__,
            context=_init_context_var.get(),
        )

    def routing_key(self):
        return f"{self.etype}:{self.eid}"

    def as_message(self) -> str:
        return self.model_dump_json(exclude_defaults=True, exclude_unset=True)

    @field_validator("etype")
    def validate_etype(cls, v, info: FieldValidationInfo):
        context = info.context
        if context and "model_spec" in context:
            assert v in context["model_spec"], f"Invalid etype '{v}'"
        else:
            raise AssertionError("Missing `model_spec` in context")
        return v


def parse_data_point_task(task: str, model_spec: ModelSpec) -> DataPointTask:
    with task_context(model_spec):
        return DataPointTask.model_validate_json(task)


class SnapshotMessageType(Enum):
    task = "task"
    linked_entities = "linked_entities"
    run_start = "run_start"
    run_end = "run_end"


class Snapshot(Task):
    """Snapshot

    Contains a list of entities, the meaning of which depends on the `type`.
    If `type` is "task", then the list contains linked entities for which a snapshot
    should be created. Otherwise `type` is "linked_entities", indicating which entities
    must be skipped in a parallelized creation of unlinked entities.

    Attributes:
        entities: List of (entity_type, entity_id)
        time: timestamp for snapshot creation
        final: If True, this is the last linked snapshot for the given time
    """

    entities: list[tuple[str, str]] = []
    time: datetime
    type: SnapshotMessageType
    final: bool = False

    def routing_key(self):
        return "-".join(f"{etype}:{eid}" for etype, eid in self.entities)

    def as_message(self) -> str:
        return self.model_dump_json()



"""
Allows modules to register functions (callables) to be run at
specified times or intervals (like cron does).

Based on APScheduler package
"""

import logging
from typing import Callable, Union

from apscheduler.schedulers.background import BackgroundScheduler
from apscheduler.triggers.cron import CronTrigger

from dp3.common.utils import get_func_name


class Scheduler:
    """
    Allows modules to register functions (callables) to be run
    at specified times or intervals (like cron does).
    """

    def __init__(self) -> None:
        self.log = logging.getLogger("Scheduler")
        # self.log.setLevel("DEBUG")
        logging.getLogger("apscheduler.scheduler").setLevel("WARNING")
        logging.getLogger("apscheduler.executors.default").setLevel("WARNING")
        self.sched = BackgroundScheduler(timezone="UTC")
        self.last_job_id = 0

    def start(self) -> None:
        self.log.debug("Scheduler start")
        self.sched.start()

    def stop(self) -> None:
        self.log.debug("Scheduler stop")
        self.sched.shutdown()

    def register(
        self,
        func: Callable,
        func_args: Union[list, tuple] = None,
        func_kwargs: dict = None,
        year: Union[int, str] = None,
        month: Union[int, str] = None,
        day: Union[int, str] = None,
        week: Union[int, str] = None,
        day_of_week: Union[int, str] = None,
        hour: Union[int, str] = None,
        minute: Union[int, str] = None,
        second: Union[int, str] = None,
        timezone: str = "UTC",
        misfire_grace_time: int = 1,
    ) -> int:
        """
        Register a function to be run at specified times.

        Pass cron-like specification of when the function should be called,
        see [docs](https://apscheduler.readthedocs.io/en/latest/modules/triggers/cron.html)
        of apscheduler.triggers.cron for details.

        Args:
            func: function or method to be called
            func_args: list of positional arguments to call func with
            func_kwargs: dict of keyword arguments to call func with
            year: 4-digit year
            month: month (1-12)
            day: day of month (1-31)
            week: ISO week (1-53)
            day_of_week: number or name of weekday (0-6 or mon,tue,wed,thu,fri,sat,sun)
            hour: hour (0-23)
            minute: minute (0-59)
            second: second (0-59)
            timezone: Timezone for time specification (default is UTC).
            misfire_grace_time: seconds after the designated run time
                that the job is still allowed to be run (default is 1)
        Returns:
             job ID
        """
        self.last_job_id += 1
        trigger = CronTrigger(
            year, month, day, week, day_of_week, hour, minute, second, timezone=timezone
        )
        self.sched.add_job(
            func,
            trigger,
            func_args,
            func_kwargs,
            coalesce=True,
            max_instances=1,
            misfire_grace_time=misfire_grace_time,
            id=str(self.last_job_id),
        )
        self.log.debug(f"Registered function {get_func_name(func)} to be called at {trigger}")
        return self.last_job_id

    def pause_job(self, id):
        """Pause job with given ID"""
        self.sched.pause_job(str(id))

    def resume_job(self, id):
        """Resume previously paused job with given ID"""
        self.sched.resume_job(str(id))



import ipaddress
import re
from datetime import datetime
from enum import Enum
from typing import Annotated, Any, Optional, Union

from pydantic import (
    BaseModel,
    Extra,
    Field,
    Json,
    PrivateAttr,
    RootModel,
    create_model,
    model_validator,
)

# Regular expressions for parsing various data types
re_array = re.compile(r"^array<(.+)>$")
re_set = re.compile(r"^set<(.+)>$")
re_link = re.compile(
    r"^link\s*<\s*(?P<etype>\w+)\s*(?:,\s*(?P<data>\S+)\s*)?(?:;\s*mirror=(?P<mirror>\S+)\s*)?>$"
)
re_dict = re.compile(r"^dict\s*<(\s*(\w+\??:\w+,\s*)*(\w+\??:\w+\s*))>$")
re_category = re.compile(r"^category<\s*(?P<type>\w+)\s*;\s*(?P<vals>(?:\s*\w+,\s*)*\s*\w+\s*)>$")

# Dictionary containing validator functions for primitive data types
primitive_data_types = {
    "tag": bool,
    "binary": bool,
    "string": str,
    "int": int,
    "int64": int,
    "float": float,
    "ipv4": ipaddress.IPv4Address,
    "ipv6": ipaddress.IPv6Address,
    "mac": Annotated[str, Field(pattern=r"^([0-9a-fA-F]{2}[:-]){5}([0-9a-fA-F]{2})$")],
    "time": datetime,
    "special": Any,
    "json": Union[Json[Any], dict, list],
}


class ReadOnly(BaseModel):
    """The ReadOnly data_type is used to avoid datapoint insertion for an attribute."""

    @model_validator(mode="before")
    @classmethod
    def fail(cls, data):
        raise ValueError("Cannot instantiate a read-only attribute datapoint.")


class Link(BaseModel, extra=Extra.forbid):
    eid: str


class DataType(RootModel):
    """Data type container

    Represents one of primitive data types:

    - tag
    - binary
    - string
    - int
    - int64
    - float
    - ipv4
    - ipv6
    - mac
    - time
    - special
    - json

    or composite data type:

    - link<entity_type>
    - array<str_type>
    - set<str_type>
    - dict<keys>
    - category<str_type;category1,category2,...>
    """

    root: str
    _data_type = PrivateAttr(None)
    _type_info = PrivateAttr(None)
    _hashable = PrivateAttr(True)

    _iterable = PrivateAttr(False)
    _elem_type = PrivateAttr()

    _is_link = PrivateAttr(False)
    _link_to = PrivateAttr()
    _mirror_link = PrivateAttr(False)
    _mirror_as = PrivateAttr()
    _link_data = PrivateAttr()

    @model_validator(mode="after")
    def determine_value_validator(self):
        """Determines value validator (inner `data_type`)."""
        str_type = self.root

        self._hashable = not (
            "dict" in str_type
            or "set" in str_type
            or "array" in str_type
            or "special" in str_type
            or "json" in str_type
            or "link" in str_type
        )

        if str_type in primitive_data_types:
            # Primitive type
            data_type = primitive_data_types[str_type]

        elif m := re.match(re_array, str_type):
            # Array
            element_type = m.group(1).strip()
            value_type = DataType(root=element_type)
            if not is_primitive_element_type(value_type):
                raise ValueError(f"Data type {element_type} is not supported as an array element")
            data_type = list[value_type.data_type]
            self._iterable = True
            self._elem_type = value_type

        elif m := re.match(re_set, str_type):
            # Set
            element_type = m.group(1).strip()
            value_type = DataType(root=element_type)
            if not is_primitive_element_type(value_type):
                raise ValueError(f"Data type {element_type} is not supported as a set element")
            data_type = list[value_type.data_type]  # set is not supported by MongoDB
            self._iterable = True
            self._elem_type = value_type

        elif m := re.match(re_link, str_type):
            # Link
            etype, data, mirrored = m.group("etype"), m.group("data"), m.group("mirror")
            self._link_to = etype
            self._is_link = True
            self._link_data = bool(data)
            self._mirror_link = bool(mirrored)
            self._mirror_as = mirrored if mirrored else None
            self._type_info = f"link<{etype},{data}>"

            if etype and data:
                value_type = DataType(root=data)
                data_type = create_model(
                    f"Link<{data}>", __base__=Link, data=(value_type.data_type, ...)
                )
            else:
                data_type = Link

        elif re.match(re_dict, str_type):
            # Dict
            dict_spec = {}

            key_str = str_type.split("<")[1].split(">")[0]
            key_spec = dict(item.strip().split(":") for item in key_str.split(","))

            # For each dict key
            for k, v in key_spec.items():
                if v not in primitive_data_types:
                    raise ValueError(f"Data type {v} of key {k} is not supported as a dict field")

                # Optional subattribute
                k_optional = k[-1] == "?"

                # Set (type, default value) for the key
                if k_optional:
                    k = k[:-1]  # Remove question mark from key
                    dict_spec[k] = (Optional[primitive_data_types[v]], None)
                else:
                    dict_spec[k] = (primitive_data_types[v], ...)

            # Create model for this dict
            data_type = create_model(f"{str_type}__inner", **dict_spec)
            self._type_info = (
                "dict<" + ",".join(f"{k}:{v}" for k, v in sorted(dict_spec.items())) + ">"
            )

        elif m := re.match(re_category, str_type):
            # Category
            category_type, category_values = m.group("type"), m.group("vals")

            category_type = DataType(root=category_type)
            category_values = [
                category_type._data_type(value.strip()) for value in category_values.split(",")
            ]

            data_type = Enum(
                f"Category<{category_type}>", {str(val): val for val in category_values}
            )
        else:
            raise ValueError(f"Data type '{str_type}' is not supported")

        # Set data type
        self._data_type = data_type
        # Set default type info
        if self._type_info is None:
            self._type_info = str(data_type)
        return self

    @property
    def data_type(self) -> Union[type, BaseModel]:
        """Type for incoming value validation"""
        return self._data_type

    @property
    def type_info(self) -> str:
        """String representation of the data type, immune to whitespace changes"""
        return self._type_info

    @property
    def hashable(self) -> bool:
        """Whether contained data is hashable"""
        return self._hashable

    @property
    def iterable(self) -> bool:
        """Whether the data type is iterable"""
        return self._iterable

    @property
    def elem_type(self) -> "DataType":
        """if `iterable`, the element data type"""
        return self._elem_type

    @property
    def is_link(self) -> bool:
        """Whether the data type is a link between entities"""
        return self._is_link

    @property
    def mirror_link(self) -> bool:
        """If `is_link`, whether the link is mirrored"""
        return self._mirror_link

    @property
    def mirror_as(self) -> Union[str, None]:
        """If `mirror_link`, what is the name of the mirrored attribute"""
        return self._mirror_as

    @property
    def link_to(self) -> str:
        """If `is_link`, the target linked entity"""
        return self._link_to

    def get_linked_entity(self) -> str:
        """Returns linked entity id. Raises ValueError if DataType is not a link."""
        try:
            return self._link_to
        except AttributeError:
            raise ValueError(f"DataType '{self}' is not a link.") from None

    def link_has_data(self) -> bool:
        """Whether link has data. Raises ValueError if DataType is not a link."""
        try:
            return self._link_data
        except AttributeError:
            raise ValueError(f"DataType '{self}' is not a link.") from None

    def __str__(self):
        return self.root

    def __repr__(self):
        return f"'{str(self)}'"


def is_primitive_element_type(data_type: DataType) -> bool:
    return data_type.root in primitive_data_types or data_type.is_link



from collections import defaultdict

from fastapi import APIRouter, Request

from dp3.api.internal.config import (
    DATAPOINTS_INGESTION_URL_PATH,
    DB,
    DP_LOGGER,
    MODEL_SPEC,
    TASK_WRITER,
)
from dp3.api.internal.entity_response_models import EntityState
from dp3.api.internal.helpers import api_to_dp3_datapoint
from dp3.api.internal.models import DataPoint
from dp3.api.internal.response_models import HealthCheckResponse, SuccessResponse
from dp3.common.task import DataPointTask, task_context

router = APIRouter()


@router.get("/", tags=["Health"])
async def health_check() -> HealthCheckResponse:
    """Health check

    Returns simple 'It works!' response.
    """
    return HealthCheckResponse()


@router.post(DATAPOINTS_INGESTION_URL_PATH, tags=["Data ingestion"])
async def insert_datapoints(dps: list[DataPoint], request: Request) -> SuccessResponse:
    """Insert datapoints

    Validates and pushes datapoints into task queue, so they are processed by one of DP3 workers.
    """
    # Convert to DP3 datapoints
    # This should not fail as all datapoints are already validated
    dp3_dps = [api_to_dp3_datapoint(dp.model_dump()) for dp in dps]

    # Group datapoints by etype-eid
    tasks_dps = defaultdict(list)
    for dp in dp3_dps:
        key = (dp.etype, dp.eid)
        tasks_dps[key].append(dp)

    # Create tasks
    tasks = []
    with task_context(MODEL_SPEC):
        for k in tasks_dps:
            etype, eid = k

            # This shouldn't fail either
            tasks.append(DataPointTask(etype=etype, eid=eid, data_points=tasks_dps[k]))

    # Push tasks to task queue
    for task in tasks:
        TASK_WRITER.put_task(task, False)

    # Log datapoints
    DP_LOGGER.log_good(dp3_dps, src=request.client.host)

    return SuccessResponse()


@router.get("/entities", tags=["Entity"])
async def list_entity_types() -> dict[str, EntityState]:
    """List entity types

    Returns dictionary containing all entity types configured -- their simplified configuration
    and current state information.
    """
    entities = {}

    for etype in MODEL_SPEC.entities:
        entity_spec = MODEL_SPEC.entity(etype)
        entities[etype] = {
            "id": etype,
            "name": entity_spec.name,
            "attribs": MODEL_SPEC.attribs(etype),
            "eid_estimate_count": DB.estimate_count_eids(etype),
        }

    return entities



"""Core module that executes periodic update callbacks."""

import logging
from collections import defaultdict
from collections.abc import Iterator
from datetime import datetime, timedelta
from functools import partial
from typing import Callable, Literal

from pydantic import BaseModel, validate_call
from pymongo.cursor import Cursor
from pymongo.results import UpdateResult

from dp3.common.config import CronExpression, PlatformConfig
from dp3.common.scheduler import Scheduler
from dp3.common.task import DataPointTask, task_context
from dp3.common.types import EventGroupType, ParsedTimedelta
from dp3.database.database import EntityDatabase
from dp3.task_processing.task_queue import TaskQueueWriter


class UpdaterConfig(BaseModel, extra="forbid"):
    """The configuration of the Updater module.

    The periodic update is executed in smaller batches for better robustness.
    The batch size is dynamically adjusted based the total number of entities
    and the estimated growth rate.

    Attributes:
        update_batch_cron: A CRON expression for the periodic update.
        update_batch_period: The period of the periodic update.
            Should equal to the period of `update_batch_cron`.
        cache_management_cron: A CRON expression for the cache management.
        cache_max_entries: The maximum number of finished cache entries per thread_id.
    """

    update_batch_cron: CronExpression
    update_batch_period: ParsedTimedelta

    cache_management_cron: CronExpression = CronExpression(hour=2, minute=0, second=0)
    cache_max_entries: int = 32


class UpdateThreadState(BaseModel, validate_assignment=True):
    """
    A cache item describing a state of one configured update thread.

    Attributes:
        type: "state"
        t_created: Time of creation.
        t_last_update: Time of last update.
        t_end: Time of predicted period end.
        processed: The number of currently processed entities.
        total: The total number of entities.
        iteration: The current iteration.
        total_iterations: Total number of iterations.
        etype: Entity type.
        period: Period length in seconds.
        eid_only: Whether only eids are passed to hooks.
        hook_ids: Hook ids.
        runtime_secs: Total hook runtime in seconds.
    """

    type: Literal["state"] = "state"
    t_created: datetime
    t_last_update: datetime
    t_end: datetime
    processed: int = 0
    total: int = 0
    iteration: int = 0
    total_iterations: int = 0
    finished: bool = False
    period: float
    etype: str
    eid_only: bool
    hook_ids: list[str]
    runtime_secs: float = 0.0

    @classmethod
    def new(cls, hooks: dict, period: float, entity_type: str, eid_only: bool = False):
        """Create a new instance initialized with hooks and thread_id components."""
        now = datetime.now()
        return cls(
            t_created=now,
            t_last_update=now,
            t_end=now + timedelta(seconds=period),
            period=period,
            etype=entity_type,
            eid_only=eid_only,
            hook_ids=hooks.keys(),
        )

    @staticmethod
    def id_attributes():
        """A list of attributes which identify the state in cache."""
        return ["type", "period", "etype", "eid_only", "t_created"]

    @property
    def thread_id(self) -> tuple[float, str, bool]:
        """A tuple of (period, entity_type, eid_only)."""
        return self.period, self.etype, self.eid_only

    def reset(self):
        """Resets counters and timestamps."""
        now = datetime.now()
        self.t_created = now
        self.t_last_update = now
        self.t_end = now + timedelta(seconds=self.period)
        self.iteration = 0
        self.processed = 0
        self.runtime_secs = 0.0
        self.finished = False


class UpdaterCache:
    """
    The cache collection contains the metadata documents with
    the state of the update process for each entity type.
    """

    def __init__(self, cache_collection):
        self._cache = cache_collection
        self._setup_cache_indexes()

    def get_unfinished(self) -> Iterator[UpdateThreadState]:
        """Yields all unfinished cache entries from DB."""
        for state in self._cache.find({"type": "state", "finished": False}):
            yield UpdateThreadState.model_validate(state)

    def upsert(self, state: UpdateThreadState) -> UpdateResult:
        """Update or insert a state entry into DB."""
        state_dict = state.model_dump()
        filter_dict = {k: v for k, v in state_dict.items() if k in state.id_attributes()}
        update_dict = {
            "$set": {k: v for k, v in state_dict.items() if k not in state.id_attributes()}
        }
        return self._cache.update_one(filter_dict, update=update_dict, upsert=True)

    def register_management(
        self, scheduler: Scheduler, trigger: CronExpression, max_entries: int
    ) -> int:
        """Registers the cache management task with the scheduler."""
        return scheduler.register(
            self._manage_cache, func_args=[max_entries], **trigger.model_dump()
        )

    def _manage_cache(self, max_entries: int):
        """Removes the oldest finished cache entries if their count exceeds the limit."""
        counts_per_thread_id = defaultdict(int)
        to_delete = []

        for state in self._cache.find({"type": "state", "finished": True}).sort("t_created", -1):
            state_obj = UpdateThreadState.model_validate(state)
            counts_per_thread_id[state_obj.thread_id] += 1
            if counts_per_thread_id[state_obj.thread_id] > max_entries:
                to_delete.append(state["_id"])

        self._cache.delete_many({"_id": {"$in": to_delete}})

    def _setup_cache_indexes(self):
        """Sets up the indexes of the state cache."""
        self._cache.create_index("t_created", background=True)


class Updater:
    """Executes periodic update callbacks."""

    def __init__(
        self,
        db: EntityDatabase,
        task_queue_writer: TaskQueueWriter,
        platform_config: PlatformConfig,
        scheduler: Scheduler,
        elog: EventGroupType,
    ):
        self.log = logging.getLogger("Updater")
        self.elog = elog

        self.model_spec = platform_config.model_spec
        self.config = UpdaterConfig.model_validate(platform_config.config.get("updater", {}))
        self.db = db
        self.task_queue_writer = task_queue_writer
        self.scheduler = scheduler

        self.enabled = platform_config.process_index == 0

        if not self.enabled:
            return

        # Get state cache
        self.cache = UpdaterCache(self.db.get_module_cache("Updater"))
        self.cache.register_management(
            scheduler, self.config.cache_management_cron, self.config.cache_max_entries
        )

        self.update_thread_hooks = defaultdict(dict)

    @validate_call
    def register_record_update_hook(
        self,
        hook: Callable[[str, str, dict], list[DataPointTask]],
        hook_id: str,
        entity_type: str,
        period: ParsedTimedelta,
    ):
        """Registers a hook for periodic update of entities of the specified type.

        The hook receives the entity type, the entity ID and the master record.
        """
        self._register_hook(hook, hook_id, entity_type, period.total_seconds(), eid_only=False)

    @validate_call
    def register_eid_update_hook(
        self,
        hook: Callable[[str, str], list[DataPointTask]],
        hook_id: str,
        entity_type: str,
        period: ParsedTimedelta,
    ):
        """Registers a hook for periodic update of entities of the specified type.

        The hook receives the entity type and the entity ID.
        """
        self._register_hook(hook, hook_id, entity_type, period.total_seconds(), eid_only=True)

    def _register_hook(self, hook, hook_id: str, entity_type: str, period: float, eid_only: bool):
        if not self.enabled:
            return

        update_period_secs = self.config.update_batch_period.total_seconds()
        if period < update_period_secs:
            raise ValueError(
                f"The total period {period}s is must be greater or equal than "
                f"the update batch period {update_period_secs}s."
            )

        thread_id = (period, entity_type, eid_only)
        hooks = self.update_thread_hooks[thread_id]
        if hook_id in hooks:
            raise ValueError(f"Hook ID {hook_id} already registered for {entity_type}.")
        self.log.info(
            "Registered hook '%s' to thread processing entity '%s' over %.1fs, eid_only = %s",
            hook_id,
            entity_type,
            period,
            eid_only,
        )
        hooks[hook_id] = hook

    def start(self):
        """
        Starts the updater.

        Will fetch the state of the updater from the cache and schedule the update threads.
        """
        if not self.enabled:
            return

        # Get all unfinished progress states
        saved_states = {}
        for state in self.cache.get_unfinished():
            saved_states[state.thread_id] = state

        # Confirm all saved states have configured hooks, terminate if not
        for thread_id, state in saved_states.items():
            if thread_id not in self.update_thread_hooks:
                self.log.warning(
                    "Previously configured hooks %s for '%s' entity with period: %ss "
                    "match no current configuration, aborting update thread.",
                    state.hook_ids,
                    state.etype,
                    state.period,
                )
                state.finished = True
                self.cache.upsert(state)
                continue

            # Find if any new hooks are added
            configured_hooks = self.update_thread_hooks[thread_id]
            saved_hook_ids = set(state.hook_ids)
            configured_hook_ids = set(configured_hooks.keys())
            new_hook_ids = configured_hook_ids - saved_hook_ids
            deleted_hook_ids = saved_hook_ids - configured_hook_ids

            # Update the state with new hooks
            if deleted_hook_ids or new_hook_ids:
                if deleted_hook_ids:
                    self.log.warning(
                        "Previously configured hooks %s were deleted for entity '%s', period: %ss",
                        deleted_hook_ids,
                        state.etype,
                        state.period,
                    )
                state.hook_ids = list(configured_hook_ids)
                self.cache.upsert(state)

        # Add newly configured hooks that are not in the saved states
        for thread_id, hooks in self.update_thread_hooks.items():
            if thread_id not in saved_states:
                state = UpdateThreadState.new(hooks, *thread_id)
                saved_states[thread_id] = state

        # Schedule the update threads
        for (period, entity_type, eid_only), state in saved_states.items():
            if state.finished:
                continue
            hooks = self.update_thread_hooks[(period, entity_type, eid_only)]
            if eid_only:
                processing_func = self._process_eid_update_batch
            else:
                processing_func = self._process_update_batch

            state.total = self.db.get_estimated_entity_count(entity_type)
            try:
                total_iterations = self._calculate_iteration_count(state.period)
                if state.iteration != 0 and total_iterations != state.total_iterations:
                    self.log.info("The update period was changed, resetting iteration number")
                    state.iteration = 0
                state.total_iterations = total_iterations
            except ValueError:
                self.log.error(
                    "Invalid period configuration for thread: %s, "
                    "the update batch period must be smaller than the total period ",
                    state,
                )
                raise

            self.scheduler.register(
                processing_func,
                func_args=[entity_type, hooks, state],
                **self.config.update_batch_cron.model_dump(),
            )

        # Connect the queue writer
        self.task_queue_writer.connect()
        self.task_queue_writer.check()  # check presence of needed exchanges

    def stop(self):
        """Stops the updater."""
        if not self.enabled:
            return

        self.task_queue_writer.disconnect()

    def _process_update_batch(self, entity_type: str, hooks: dict, state: UpdateThreadState):
        """Processes a batch of entities of the specified type.

        Args:
            entity_type: The entity type.
            hooks: The update hooks.
            state: The state of the update process.
        """
        self._process_batch(
            entity_type,
            hooks,
            state,
            record_getter=self.db.get_worker_master_records,
            hook_runner=self._run_hooks,
        )

    def _process_eid_update_batch(self, entity_type: str, hooks: dict, state: UpdateThreadState):
        """Processes a batch of entities of the specified type, only passing the entity ID.

        Args:
            entity_type: The entity type.
            hooks: The update hooks.
            state: The state of the update process.
        """
        projection = {"_id": True}
        self._process_batch(
            entity_type,
            hooks,
            state,
            record_getter=partial(self.db.get_worker_master_records, projection=projection),
            hook_runner=self._run_hooks_eid,
        )

    def _process_batch(
        self,
        entity_type: str,
        hooks: dict,
        state: UpdateThreadState,
        record_getter: Callable[[int, int, str], Cursor],
        hook_runner: Callable[[dict[str, Callable], str, dict], None],
    ):
        """Processes a batch of entities of the specified type using the specified `value_getter`.

        Args:
            entity_type: The entity type.
            hooks: The update hooks.
            state: The state of the update process.
            record_getter: Callable taking the (iteration, total_iterations, entity_type)
                to access the required record values.
            hook_runner: Callable taking the (hooks, etype, record) and running the hooks.
        """
        self.log.debug(
            "Processing update batch for '%s' over %.1fs, eid_only = %s",
            entity_type,
            state.period,
            state.eid_only,
        )
        start = datetime.now()
        iteration_cnt = state.total_iterations
        iteration = state.iteration

        self.log.debug(
            "Current state - processed entities: %s/%s, iteration: %s/%s, "
            "avg runtime per entity: %.3fs",
            state.processed,
            state.total,
            iteration,
            state.total_iterations,
            state.runtime_secs / state.processed if state.processed > 0 else 0,
        )
        records = record_getter(iteration, iteration_cnt, entity_type)

        for record in records:
            hook_runner(hooks, entity_type, record)

            state.processed += 1
        state.t_last_update = datetime.now()

        duration = state.t_last_update - start
        state.runtime_secs += duration.total_seconds()

        if state.runtime_secs > state.period:
            self.log.warning(
                "%s hooks: %s, The total hook runtime %.2fs exceeds the desired period %.2fs. "
                "(Current average is %.3fs per entity when limit is %.3fs) "
                "Consider optimizing the hooks or extending the update period.",
                entity_type,
                state.hook_ids,
                state.runtime_secs,
                state.period,
                state.runtime_secs / state.processed,
                state.period / state.total,
            )

        state.iteration = iteration + 1
        if state.iteration >= iteration_cnt:
            state.finished = True

        self.cache.upsert(state)

        if state.finished:
            self.log.debug(
                "Finished processing '%s' entity with period: %s", entity_type, state.period
            )
            state.reset()
            state.total = self.db.get_estimated_entity_count(entity_type)

    def _calculate_iteration_count(self, period_seconds: float) -> int:
        total_iterations = int(period_seconds // self.config.update_batch_period.total_seconds())
        if total_iterations == 0:
            raise ValueError("The total period is shorter than the update batch period.")
        return total_iterations

    def _run_hooks(self, hooks: dict[str, Callable], entity_type: str, record: dict):
        tasks = []
        with task_context(self.model_spec):
            for hook_id, hook in hooks.items():
                self.log.debug("Running hook: '%s'", hook_id)
                try:
                    new_tasks = hook(entity_type, record["_id"], record)
                    tasks.extend(new_tasks)
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")

        for task in tasks:
            self.task_queue_writer.put_task(task)

    def _run_hooks_eid(self, hooks: dict[str, Callable], entity_type: str, record: dict):
        tasks = []
        with task_context(self.model_spec):
            for hook_id, hook in hooks.items():
                self.log.debug("Running hook: '%s'", hook_id)
                try:
                    new_tasks = hook(entity_type, record["_id"])
                    tasks.extend(new_tasks)
                except Exception as e:
                    self.elog.log("module_error")
                    self.log.error(f"Error during running hook {hook}: {e}")

        for task in tasks:
            self.task_queue_writer.put_task(task)
